from __future__ import print_function
from __future__ import absolute_import
import numpy
from .Sed import Sed
from .Bandpass import Bandpass
from . import LSSTdefaults

__all__ = ["FWHMeff2FWHMgeom", "FWHMgeom2FWHMeff",
           "calcNeff", "calcInstrNoiseSq", "calcTotalNonSourceNoiseSq", "calcSNR_sed",
          "calcM5", "calcSkyCountsPerPixelForM5", "calcGamma", "calcSNR_m5",
          "calcAstrometricError", "magErrorFromSNR", "calcMagError_m5", "calcMagError_sed"]

def FWHMeff2FWHMgeom(FWHMeff):
    """
    Convert FWHMeff to FWHMgeom.
    This conversion was calculated by Bo Xin and Zeljko Ivezic (and will be in an update on the LSE-40 and overview papers).
    @param [in] FWHMeff (the single-gaussian equivalent FWHM value, appropriate for calcNeff) in arcseconds
    @param [out] FWHMgeom (the geometric FWHM value, as measured from a typical PSF profile) in arcseconds
    """
    FWHMgeom = 0.822*FWHMeff + 0.052
    return FWHMgeom

def FWHMgeom2FWHMeff(FWHMgeom):
    """
    Convert FWHMgeom to FWHMeff.
    This conversion was calculated by Bo Xin and Zeljko Ivezic (and will be in an update on the LSE-40 and overview papers).
    @param [in] FWHMgeom (the geometric FWHM value, as measured from a typical PSF profile) in arcseconds
    @param [out] FWHMeff (the single-gaussian equivalent FWHM value, appropriate for calcNeff) in arcseconds
    """
    FWHMeff = (FWHMgeom - 0.052)/0.822
    return FWHMeff

def calcNeff(FWHMeff, platescale):
    """
    Calculate the effective number of pixels in a single gaussian PSF.
    This equation comes from LSE-40, equation 27.
    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40
    @param [in] FWHMeff in arcseconds
       (the width of a single-gaussian that produces correct Neff for typical PSF profile)
    @param [in] platescale in arcseconds per pixel
    @param [out] the effective number of pixels contained in the PSF
    The FWHMeff is a way to represent the equivalent seeing value, if the
    atmosphere could be simply represented as a single gaussian (instead of a more
    complicated von Karman profile for the atmosphere, convolved properly with the
    telescope hardware additional blurring of 0.4").
    A translation from the geometric FWHM to the FWHMeff is provided in FWHMgeom2FWHMeff.
    """
    return 2.266*(FWHMeff/platescale)**2


def calcInstrNoiseSq(photParams):
    """
    Combine all of the noise due to intrumentation into one value
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [out] The noise due to all of these sources added in quadrature
    in ADU counts
    """

    # instrumental squared noise in electrons
    instNoiseSq = photParams.nexp*photParams.readnoise**2 + \
                  photParams.darkcurrent*photParams.exptime*photParams.nexp + \
                  photParams.nexp*photParams.othernoise**2

    # convert to ADU counts
    instNoiseSq = instNoiseSq/(photParams.gain*photParams.gain)

    return instNoiseSq


def calcTotalNonSourceNoiseSq(skySed, hardwarebandpass, photParams, FWHMeff):
    """
    Calculate the noise due to things that are not the source being observed
    (i.e. intrumentation and sky background)
    @param [in] skySed -- an instantiation of the Sed class representing the sky
    (normalized so that skySed.calcMag() gives the sky brightness in magnitudes
    per square arcsecond)
    @param [in] hardwarebandpass -- an instantiation of the Bandpass class representing
    just the instrumentation throughputs
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [in] FWHMeff in arcseconds
    @param [out] total non-source noise squared (in ADU counts)
    (this is simga^2_tot * neff in equation 41 of the SNR document
    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40 )
    """

    # Calculate the effective number of pixels for double-Gaussian PSF
    neff = calcNeff(FWHMeff, photParams.platescale)

    # Calculate the counts from the sky.
    # We multiply by two factors of the platescale because we expect the
    # skySed to be normalized such that calcADU gives counts per
    # square arc second, and we need to convert to counts per pixel.

    skycounts = skySed.calcADU(hardwarebandpass, photParams=photParams) \
                * photParams.platescale * photParams.platescale

    # Calculate the square of the noise due to instrumental effects.
    # Include the readout noise as many times as there are exposures

    noise_instr_sq = calcInstrNoiseSq(photParams=photParams)

    # Calculate the square of the noise due to sky background poisson noise
    noise_sky_sq = skycounts/photParams.gain

    # Discount error in sky measurement for now
    noise_skymeasurement_sq = 0

    total_noise_sq = neff*(noise_sky_sq + noise_instr_sq + noise_skymeasurement_sq)

    return total_noise_sq


def calcSkyCountsPerPixelForM5(m5target, totalBandpass, photParams, FWHMeff=None):
    """
    Calculate the number of sky counts per pixel expected for a given
    value of the 5-sigma limiting magnitude (m5)
    The 5-sigma limiting magnitude (m5) for an observation is
    determined by a combination of the telescope and camera parameters
    (such as diameter of the mirrors and the readnoise) together with the
    sky background.
    @param [in] the desired value of m5
    @param [in] totalBandpass is an instantiation of the Bandpass class
    representing the total throughput of the telescope (instrumentation
    plus atmosphere)
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [in] FWHMeff in arcseconds
    @param [out] returns the expected number of sky counts per pixel
    """

    if FWHMeff is None:
        FWHMeff = LSSTdefaults().FWHMeff('r')

    # instantiate a flat SED
    flatSed = Sed()
    flatSed.setFlatSED()

    # normalize the SED so that it has a magnitude equal to the desired m5
    fNorm = flatSed.calcFluxNorm(m5target, totalBandpass)
    flatSed.multiplyFluxNorm(fNorm)
    sourceCounts = flatSed.calcADU(totalBandpass, photParams=photParams)

    # calculate the effective number of pixels for a double-Gaussian PSF
    neff = calcNeff(FWHMeff, photParams.platescale)

    # calculate the square of the noise due to the instrument
    noise_instr_sq = calcInstrNoiseSq(photParams=photParams)

    # now solve equation 41 of the SNR document for the neff * sigma_total^2 term
    # given snr=5 and counts as calculated above
    # SNR document can be found at
    # https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40

    nSigmaSq = (sourceCounts*sourceCounts)/25.0 - sourceCounts/photParams.gain

    skyNoiseTarget = nSigmaSq/neff - noise_instr_sq
    skyCountsTarget = skyNoiseTarget*photParams.gain

    # TODO:
    # This method should throw an error if skyCountsTarget is negative
    # unfortunately, that currently happens for default values of
    # m5 as taken from arXiv:0805.2366, table 2.  Adding the error
    # should probably wait for a later issue in which we hash out what
    # the units are for all of the parameters stored in PhotometricDefaults.

    return skyCountsTarget


def calcM5(skysed, totalBandpass, hardware, photParams, FWHMeff=None):
    """
    Calculate the AB magnitude of a 5-sigma above sky background source.
    The 5-sigma limiting magnitude (m5) for an observation is determined by
    a combination of the telescope and camera parameters (such as diameter
    of the mirrors and the readnoise) together with the sky background. This
    method (calcM5) calculates the expected m5 value for an observation given
    a sky background Sed and hardware parameters.
    @param [in] skysed is an instantiation of the Sed class representing
    sky emission, normalized so that skysed.calcMag gives the sky brightness
    in magnitudes per square arcsecond.
    @param [in] totalBandpass is an instantiation of the Bandpass class
    representing the total throughput of the telescope (instrumentation
    plus atmosphere)
    @param [in] hardware is an instantiation of the Bandpass class representing
    the throughput due solely to instrumentation.
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [in] FWHMeff in arcseconds
    @param [out] returns the value of m5 for the given bandpass and sky SED
    """
    # This comes from equation 45 of the SNR document (v1.2, May 2010)
    # https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40

    if FWHMeff is None:
        FWHMeff = LSSTdefaults().FWHMeff('r')

    # create a flat fnu source
    flatsource = Sed()
    flatsource.setFlatSED()
    snr = 5.0
    v_n = calcTotalNonSourceNoiseSq(skysed, hardware, photParams, FWHMeff)

    counts_5sigma = (snr**2)/2.0/photParams.gain + \
                     numpy.sqrt((snr**4)/4.0/photParams.gain + (snr**2)*v_n)

    # renormalize flatsource so that it has the required counts to be a 5-sigma detection
    # given the specified background
    counts_flat = flatsource.calcADU(totalBandpass, photParams=photParams)
    flatsource.multiplyFluxNorm(counts_5sigma/counts_flat)

    # Calculate the AB magnitude of this source.
    mag_5sigma = flatsource.calcMag(totalBandpass)
    return mag_5sigma


def magErrorFromSNR(snr):
    """
    convert flux signal to noise ratio to an error in magnitude
    @param [in] snr is the signal to noise ratio in flux
    @param [out] the resulting error in magnitude
    """

    #see www.ucolick.org/~bolte/AY257/s_n.pdf section 3.1
    return 2.5*numpy.log10(1.0+1.0/snr)


def calcGamma(bandpass, m5, photParams):

    """
    Calculate the gamma parameter used for determining photometric
    signal to noise in equation 5 of the LSST overview paper
    (arXiv:0805.2366)
    @param [in] bandpass is an instantiation of the Bandpass class
    representing the bandpass for which you desire to calculate the
    gamma parameter
    @param [in] m5 is the magnitude at which a 5-sigma detection occurs
    in this Bandpass
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [out] gamma
    """
    # This is based on the LSST SNR document (v1.2, May 2010)
    # https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40
    # as well as equations 4-6 of the overview paper (arXiv:0805.2366)

    # instantiate a flat SED
    flatSed = Sed()
    flatSed.setFlatSED()

    # normalize the SED so that it has a magnitude equal to the desired m5
    fNorm = flatSed.calcFluxNorm(m5, bandpass)
    flatSed.multiplyFluxNorm(fNorm)
    counts = flatSed.calcADU(bandpass, photParams=photParams)

    # The expression for gamma below comes from:
    #
    # 1) Take the approximation N^2 = N0^2 + alpha S from footnote 88 in the overview paper
    # where N is the noise in flux of a source, N0 is the noise in flux due to sky brightness
    # and instrumentation, S is the number of counts registered from the source and alpha
    # is some constant
    #
    # 2) Divide by S^2 and demand that N/S = 0.2 for a source detected at m5. Solve
    # the resulting equation for alpha in terms of N0 and S5 (the number of counts from
    # a source at m5)
    #
    # 3) Substitute this expression for alpha back into the equation for (N/S)^2
    # for a general source.  Re-factor the equation so that it looks like equation
    # 5 of the overview paper (note that x = S5/S).  This should give you gamma = (N0/S5)^2
    #
    # 4) Solve equation 41 of the SNR document for the neff * sigma_total^2 term
    # given snr=5 and counts as calculated above.  Note that neff * sigma_total^2
    # is N0^2 in the equation above
    #
    # This should give you

    gamma = 0.04 - 1.0/(counts*photParams.gain)

    return gamma


def calcSNR_m5(magnitude, bandpass, m5, photParams, gamma=None):
    """
    Calculate signal to noise in flux using the model from equation (5) of arXiv:0805.2366
    @param [in] magnitude of the sources whose signal to noise you are calculating
    (can be a numpy array)
    @param [in] bandpass (an instantiation of the class Bandpass) in which the magnitude
    was calculated
    @param [in] m5 is the 5-sigma limiting magnitude for the bandpass
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [in] gamma (optional) is the gamma parameter from equation(5) of
    arXiv:0805.2366.  If not provided, this method will calculate it.
    @param [out] snr is the signal to noise ratio corresponding to
    the input magnitude.
    @param [out] gamma is  the calculated gamma parameter for the
    bandpass used here (in case the user wants to call this method again).
    Note: You can also pass in a numpy array of magnitudes calculated
    in the same bandpass with the same m5 and get a numpy array of SNR out.
    """

    if gamma is None:
        gamma = calcGamma(bandpass, m5, photParams=photParams)

    dummySed = Sed()
    m5Flux = dummySed.fluxFromMag(m5)
    sourceFlux = dummySed.fluxFromMag(magnitude)

    fluxRatio = m5Flux/sourceFlux

    noise = numpy.sqrt((0.04-gamma)*fluxRatio+gamma*fluxRatio*fluxRatio)

    return 1.0/noise, gamma


def calcMagError_m5(magnitude, bandpass, m5, photParams, gamma=None):
    """
    Calculate magnitude error using the model from equation (5) of arXiv:0805.2366
    @param [in] magnitude of the source whose error you want
    to calculate (can be a numpy array)
    @param [in] bandpass (an instantiation of the Bandpass class) in question
    @param [in] m5 is the 5-sigma limiting magnitude in that bandpass
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [in] gamma (optional) is the gamma parameter from equation(5) of
    arXiv:0805.2366.  If not provided, this method will calculate it.
    @param [out] the error associated with the magnitude
    @param [out] gamma is  the calculated gamma parameter for the
    bandpass used here (in case the user wants to call this method again).
    Note: you can also pass in a numpy of array of magnitudes calculated in
    the same Bandpass with the same m5 and get a numpy array of errors out.
    """

    snr, gamma = calcSNR_m5(magnitude, bandpass, m5, photParams, gamma=gamma)

    if photParams.sigmaSys is not None:
        return numpy.sqrt(numpy.power(magErrorFromSNR(snr),2) + numpy.power(photParams.sigmaSys,2)), gamma
    else:
        return magErrorFromSNR(snr), gamma


def calcSNR_sed(sourceSed, totalbandpass, skysed, hardwarebandpass,
                    photParams, FWHMeff, verbose=False):
    """
    Calculate the signal to noise ratio for a source, given the bandpass(es) and sky SED.
    For a given source, sky sed, total bandpass and hardware bandpass, as well as
    FWHMeff / exptime, calculates the SNR with optimal PSF extraction
    assuming a double-gaussian PSF.
    @param [in] sourceSed is an instantiation of the Sed class containing the SED of
    the object whose signal to noise ratio is being calculated
    @param [in] totalbandpass is an instantiation of the Bandpass class
    representing the total throughput (system + atmosphere)
    @param [in] skysed is an instantiation of the Sed class representing
    the sky emission per square arcsecond.
    @param [in] hardwarebandpass is an instantiation of the Bandpass class
    representing just the throughput of the system hardware.
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [in] FWHMeff in arcseconds
    @param [in] verbose is a boolean
    @param [out] signal to noise ratio
    """

    # Calculate the counts from the source.
    sourcecounts = sourceSed.calcADU(totalbandpass, photParams=photParams)

    # Calculate the (square of the) noise due to signal poisson noise.
    noise_source_sq = sourcecounts/photParams.gain

    non_source_noise_sq = calcTotalNonSourceNoiseSq(skysed, hardwarebandpass, photParams, FWHMeff)

    # Calculate total noise
    noise = numpy.sqrt(noise_source_sq + non_source_noise_sq)
    # Calculate the signal to noise ratio.
    snr = sourcecounts / noise
    if verbose:
        skycounts = skysed.calcADU(hardwarebandpass, photParams) * (photParams.platescale**2)
        noise_sky_sq = skycounts/photParams.gain
        neff = calcNeff(FWHMeff, photParams.platescale)
        noise_instr_sq = calcInstrNoiseSq(photParams)

        print("For Nexp %.1f of time %.1f: " % (photParams.nexp, photParams.exptime))
        print("Counts from source: %.2f  Counts from sky: %.2f" %(sourcecounts, skycounts))
        print("FWHMeff: %.2f('')  Neff pixels: %.3f(pix)" %(FWHMeff, neff))
        print("Noise from sky: %.2f Noise from instrument: %.2f" \
            %(numpy.sqrt(noise_sky_sq), numpy.sqrt(noise_instr_sq)))
        print("Noise from source: %.2f" %(numpy.sqrt(noise_source_sq)))
        print(" Total Signal: %.2f   Total Noise: %.2f    SNR: %.2f" %(sourcecounts, noise, snr))
        # Return the signal to noise value.
    return snr


def calcMagError_sed(sourceSed, totalbandpass, skysed, hardwarebandpass,
                    photParams, FWHMeff, verbose=False):
    """
    Calculate the magnitudeError for a source, given the bandpass(es) and sky SED.
    For a given source, sky sed, total bandpass and hardware bandpass, as well as
    FWHMeff / exptime, calculates the SNR with optimal PSF extraction
    assuming a double-gaussian PSF.
    @param [in] sourceSed is an instantiation of the Sed class containing the SED of
    the object whose signal to noise ratio is being calculated
    @param [in] totalbandpass is an instantiation of the Bandpass class
    representing the total throughput (system + atmosphere)
    @param [in] skysed is an instantiation of the Sed class representing
    the sky emission per square arcsecond.
    @param [in] hardwarebandpass is an instantiation of the Bandpass class
    representing just the throughput of the system hardware.
    @param [in] photParams is an instantiation of the
    PhotometricParameters class that carries details about the
    photometric response of the telescope.
    @param [in] FWHMeff in arcseconds
    @param [in] verbose is a boolean
    @param [out] magnitude error
    """

    snr = calcSNR_sed(sourceSed, totalbandpass, skysed, hardwarebandpass,
                      photParams, FWHMeff, verbose=verbose)

    if photParams.sigmaSys is not None:
        return numpy.sqrt(numpy.power(magErrorFromSNR(snr),2) + numpy.power(photParams.sigmaSys,2))
    else:
        return magErrorFromSNR(snr)


def calcAstrometricError(mag, m5, nvisit=1):
    """
    Calculate the astrometric error, for object catalog purposes.
    Returns astrometric error for a given SNR, in mas.
    """
    # The astrometric error can be applied to parallax or proper motion (for nvisit>1).
    # If applying to proper motion, should also divide by the # of years of the survey.
    # This is also referenced in the astroph/0805.2366 paper.
    # D. Monet suggests sqrt(Nvisit/2) for first 3 years, sqrt(N) for longer, in reduction of error
    # because of the astrometric measurement method, the systematic and random error are both reduced.
    # Zeljko says 'be conservative', so removing this reduction for now.
    rgamma = 0.039
    xval = numpy.power(10, 0.4*(mag-m5))
    # The average FWHMeff is 0.7" (or 700 mas).
    error_rand = 700.0 * numpy.sqrt((0.04-rgamma)*xval + rgamma*xval*xval)
    error_rand = error_rand / numpy.sqrt(nvisit)
    # The systematic error floor in astrometry:
    error_sys = 10.0
    # These next few lines are the code removed due to Zeljko's 'be conservative' requirement.
    #if (nvisit<30):
    #    error_sys = error_sys/numpy.sqrt(nvisit/2.0)
    #if (nvisit>30):
    #    error_sys = error_sys/numpy.sqrt(nvisit)
    astrom_error = numpy.sqrt(error_sys * error_sys + error_rand*error_rand)
    return astrom_error


# pylint: skip-file
from builtins import object
__all__ = ["PhysicalParameters"]

class PhysicalParameters(object):
    """
    A class to store physical constants and other immutable parameters
    used by the sims_photUtils code
    """

    def __init__(self):
        #the quantities below are in nanometers
        self._minwavelen = 300.0
        self._maxwavelen = 1150.0
        self._wavelenstep = 0.1

        self._lightspeed = 299792458.0      # speed of light, = 2.9979e8 m/s
        self._planck = 6.626068e-27        # planck's constant, = 6.626068e-27 ergs*seconds
        self._nm2m = 1.00e-9               # nanometers to meters conversion = 1e-9 m/nm
        self._ergsetc2jansky = 1.00e23     # erg/cm2/s/Hz to Jansky units (fnu)

    @property
    def minwavelen(self):
        """
        minimum wavelength in nanometers
        """
        return self._minwavelen

    @minwavelen.setter
    def minwavelen(self, value):
        raise RuntimeError('Cannot change the value of minwavelen')


    @property
    def maxwavelen(self):
        """
        maximum wavelength in nanometers
        """
        return self._maxwavelen

    @maxwavelen.setter
    def maxwavelen(self, value):
        raise RuntimeError('Cannot change the value of maxwavelen')


    @property
    def wavelenstep(self):
        """
        wavelength step in nanometers
        """
        return self._wavelenstep

    @wavelenstep.setter
    def wavelenstep(self, value):
        raise RuntimeError('Cannot change the value of wavelenstep')


    @property
    def lightspeed(self):
        """
        speed of light in meters per second
        """
        return self._lightspeed

    @lightspeed.setter
    def lightspeed(self, value):
        raise RuntimeError('Cannot change the value of lightspeed ' +
                           '(Einstein does not approve)')


    @property
    def nm2m(self):
        """
        conversion factor to go from nm to m
        """
        return self._nm2m

    @nm2m.setter
    def nm2m(self, value):
        raise RuntimeError('Cannot change the value of nm2m')


    @property
    def ergsetc2jansky(self):
        """
        conversion factor to go from ergs/sec/cm^2 to Janskys
        """
        return self._ergsetc2jansky

    @ergsetc2jansky.setter
    def ergsetc2jansky(self, value):
        raise RuntimeError('Cannot change the value of ergsetc2Jansky')


    @property
    def planck(self):
        """
        Planck's constant in ergs*seconds
        """
        return self._planck

    @planck.setter
    def planck(self, value):
        raise RuntimeError('Cannot change the value of planck')



from builtins import object
import numpy

__all__ = ["LSSTdefaults"]

class LSSTdefaults(object):
    """
    This class exists to store default values of seeing, m5, and gamma taken from the over
    view paper (arXiv 0805.2366, Table 2, 29 August 2014 version)
    """

    def __init__(self):
        # Standard FWHMeffective in arcseconds
        self._FWHMeff = {'u':0.92, 'g':0.87, 'r':0.83, 'i':0.80, 'z':0.78, 'y':0.76}
        # Expected effective wavelength for throughput curves, in nanometers
        self._effwavelen = {'u':367.0, 'g':482.5, 'r':622.2, 'i':754.5, 'z':869.1, 'y':971.0}
        # Expected m5 depths (using FWHMeffective + dark sky + X=1.2 atmosphere + throughput curves)
        self._m5 = {'u':23.68, 'g':24.89, 'r':24.43, 'i':24.00, 'z':24.45, 'y':22.60}
        self._gamma = {'u':0.037, 'g':0.038, 'r':0.039, 'i':0.039, 'z':0.040, 'y':0.040}


    def m5(self, tag):
        """
        From arXiv 0805.2366  (Table 2):
        Typical 5-sigma depth for point sources at zenith, assuming
        exposure time of 2 x 15 seconds and observing conditions as listed.
        Calculated using $SYSENG_THROUGHPUT curves as of 11/25/2015, using
        $SYSENG_THROUGHPUT/python/calcM5.py
        @param [in] the name of a filter i.e. 'u', 'g', 'r', 'i', 'z', or 'y'
        @param [out] the corresponding m5 value
        """
        return self._m5[tag]


    def FWHMeff(self, tag):
        """
        From arXiv 0805.2366 XXX version (Table 2):
        The expected FWHMeff in arcseconds. This is the width of a single gaussian
        which produces the appropriate number of effective pixels in the PSF (thus 'FWHMeff').
        This is the value to use for calculating Neffective, when Neffective assumes a single gaussian.
        It can be converted to a geometric FWHM (equivalent to the approximate value which would
        be measured across a van Karmen PSF profile) using SignalToNoise.FWHMeff2FWHMgeom.
        @param [in] the name of a filter i.e. 'u', 'g', 'r', 'i', 'z', or 'y'
        @param [out] the corresponding FWHMeff
        """

        return self._FWHMeff[tag]

    def effwavelen(self, tag):
        """
        From the throughput curves in syseng_throughputs, calculated by
        $SYSENG_THROUGHPUTS/python/effectiveWavelen.py
        as of 11/25/2015.
        """
        return self._effwavelen[tag]


    def gamma(self, tag):
        """
        See Table 2 and Equaiton 5 of arXiv 0805.2366 29 August 2014 version.
        @param [in] the name of a filter i.e. 'u', 'g', 'r', 'i', 'z', or 'y'
        @param [out] the corresponding value of gamma as defined in the
        reference above
        """

        return self._gamma[tag]


__all__ = ["Bandpass", "BandpassDict", "PhysicalParameters", "Sed"]

from .Bandpass import Bandpass
from .BandpassDict import BandpassDict
from .PhysicalParameters import PhysicalParameters
from .Sed import Sed
from .PhotometricParameters import PhotometricParameters
from .SignalToNoise import *


import os
import numpy as np
import multiprocessing as mp
import pickle
from skopt import gp_minimize, Optimizer
from skopt.space import Real
from sklearn.externals.joblib import Parallel, delayed
from copy import deepcopy
from functools import reduce
from . import filters, spectra, calcIG, _siggiBase
from .lsst_utils import BandpassDict

__all__ = ["siggi"]


def unwrap_self_f(arg, arg2, **kwarg):
    return siggi.calc_results(arg, arg2, **kwarg)


class siggi(_siggiBase):

    """
    Class to run a complete siggi maximization run.

    Input
    -----

    spec_list, list

        List of spectra you wish to use as templates
        already stored as LSST Sims SED Objects

    spec_weights, list

        The weight you wish to assign to each SED
        template, usually just 1./len(spec_list)

    z_prior, function

        The redshift prior function you wish to use.

    z_min, float

        The minimum redshift for each SED

    z_max, float

        The maximum redshift for each SED

    z_steps, int

        The number of steps in the redshift grid

    calib_filter, Bandpass Object, default = lsst 'r' band

        The filter you wish to use to set the sky brightness
        and brightness of the SEDs when optimizing.
    """

    def __init__(self, spec_list, spec_weights, z_prior,
                 z_min=0.05, z_max=2.5, z_steps=50,
                 calib_filter=None):

        self.shift_seds = []
        self.z_probs = []

        for spec, weight in zip(spec_list, spec_weights):
            for z_val in np.linspace(z_min, z_max, z_steps):
                spec_copy = deepcopy(spec)
                spec_copy.redshiftSED(z_val)
                self.shift_seds.append(spec_copy)
                self.z_probs.append(z_prior(z_val)*weight)

        bp_dict_folder = os.path.join(os.path.dirname(__file__),
                                      'data',
                                      'lsst_baseline_throughputs')
        bp_dict = BandpassDict.loadTotalBandpassesFromFiles(
            bandpassDir=bp_dict_folder)

        if calib_filter is None:
            self.calib_filter = bp_dict['r']
        else:
            self.calib_filter = calib_filter

    def optimize_filters(self, filt_min=300., filt_max=1100.,
                         sky_mag=19.0, sed_mags=22.0, num_filters=6,
                         filter_type='trap', frozen_filt_dict=None,
                         frozen_filt_eff_wavelen=None,
                         set_ratio=None,
                         width_min=30., width_max=120.,
                         starting_points=None,
                         system_wavelen_min=300., system_wavelen_max=1150.,
                         system_wavelen_step=0.1,
                         procs=1, n_opt_points=100, acq_func_kwargs_dict=None,
                         acq_opt_kwargs_dict=None, checkpointing=True,
                         optimizer_verbosity=5,
                         parallel_backend="multiprocessing",
                         max_search_factor=50, rand_state=None,
                         load_optimizer=None,
                         save_optimizer=None):

        """
        Run the optimization for redshift estimation over the input filters
        and redshift grid. Define the brightness of the galaxies and the sky.

        Input
        -----

        filt_min, float, default = 300.

            The minimum wavelength in nm of the bandpasses.

        filt_max, float, default = 1100.

            The maximum wavelength in nm of the bandpasses.

        sky_mag, float, default = 19.0

            The default magnitude of the sky background in the calib
            filter specified when instantiating.

        sed_mags, float, default = 22.0

            The default magnitude of the SEDs in the calib filter
            specified when instantiating.

        num_filters, int, default = 6

            The number of filters to optimize

        filter_type, str, default = 'trap'

            The shape of the filters. Currently can only create
            trapezoidal shapes before the CCD response function
            is included.

        frozen_filt_dict, BandpassDict Object or None, default = None

            A set of bandpasses to include in the calculation of colors
            but not to be changed in the optimization. These will stay
            in place while other filters are optimized. This should be
            the bandpass files before the hardware and atmosphere effects
            are included in the throughput.

        frozen_filt_eff_wavelen, list of floats or None, default = None

            If a set of frozen filters is included then a set of
            effective wavelengths must also be added in order to sort their
            positions when calculating colors.

        set_ratio, float or None, default = None

            Use this to only optimize over the width of a set of trapezoidal
            filters with a set ratio between the top width and the bottom width
            specified here. For instance, 1.0 will mean a top hat filter.

        width_min, float, default = 30.

            If set_ratio is used then this is the minimum width the bottom
            of the filter must have.

        width_max, float, default = 120.

            If set_ratio is used then this is the maximum width the bottom
            of the filter can have.

        starting_points, list (n_pts, n_filters, 2 or 4) or None,
            default = None

            This is a list of starting point for the values of the filter
            corners. This is a list of lists where each list contains the s
            tarting positions of the filters.

            If set_ratio is not None then
            each of these lists have two floats to specify the bottom left
            and bottom right of a filter.

            If set_ratio is None then this is the same
            except that the lists are four numbers where the order is bottom
            left, top left, top right, bottom right corners of the filters.

        system_wavelen_min, float, default = 300.

            This is the minimum edge of the Bandpass Objects in nm.

        system_wavelen_max, float default = 1150.

            This is the maximum edge of the Bandpass Objects in nm.

        system_wavelen_step, float default = 0.1

            This is the wavelength step of the Bandpass Objects in nm.

        procs, int, default = 1

            This is the number of processors to use in optimization.

        n_opt_points, int, default = 100

            The minimum number of points to calculate the information gain
            in the optimization. If this is not a multiple of the
            number of processors used then the actual number will be
            greater than this.

        max_search_factor, int, default = 50

            This number multiplied by the number of processors used is the
            maximum number of points the optimizer will try to find
            allowable filter edges before a random allowable set of numbers
            will be chosen to fill in the remaining points in a search
            iteration. This happens since the optimizer will try to pick points
            where the filters are not in order of lowest filter center to
            highest at the beginning of the optimization
            and we want to restrict the parameter space to only where
            this is the case in order to better optimize in shorter time.

        rand_state, numpy RandomState, int or None, default = None

            Provide a RandomState Object or the integer seed for a RandomState
            in order to generate reproducible results.

        load_optimizer, None or scikit-optimizer optimizer object,
            default = None

            If this is a previously saved object then the optimizer will
            continue from this state for n_opt_points. This assumes the
            saved optimizer state ran for at least 10 points.

        save_optimizer, None or str, default = None

            If this is a string then it will save a copy of the state of the
            optimizer at the end to the filename given by the string.

        Returns
        -------

        opt, Optimizer object

            Contains the input and output values of the optimization. Use the
            absolute value of opt.yi for the information gain values. The
            corresponding filter inputs are in opt.Xi.
        """

        self.num_filters = num_filters
        self.ratio = set_ratio
        self.sky_mag = sky_mag
        self.sed_mags = sed_mags
        self.f = filters(system_wavelen_min,
                         system_wavelen_max,
                         system_wavelen_step)
        self.frozen_filt_dict = frozen_filt_dict
        self.frozen_eff_lambda = frozen_filt_eff_wavelen
        self.verbosity = optimizer_verbosity

        if type(rand_state) is int:
            rand_state = np.random.RandomState(rand_state)

        dim_list, x0 = self.set_starting_points(starting_points,
                                                self.num_filters,
                                                filt_min,
                                                filt_max,
                                                ratio=self.ratio,
                                                rand_state=rand_state)
        if self.verbosity >= 10:
            print(dim_list, x0)

        i = 0
        random_points_used = 0

        if load_optimizer is None:
            opt = Optimizer(dimensions=[Real(x1, x2) for x1, x2 in dim_list],
                            random_state=rand_state,
                            acq_func_kwargs=acq_func_kwargs_dict,
                            acq_optimizer_kwargs=acq_opt_kwargs_dict)
        else:
            opt = load_optimizer

        with Parallel(n_jobs=procs, batch_size=1, backend=parallel_backend,
                      verbose=self.verbosity) as parallel:
            while i < n_opt_points:
                if ((i == 0) and (load_optimizer is None)):
                    x = x0
                else:
                    x = []
                    pts_needed = procs
                    pts_tried = 0
                    while len(x) < procs:
                        x_pot = opt.ask(n_points=pts_needed, strategy='cl_max')
                        for point in x_pot:
                            filt_input = \
                              self.validate_filter_input(point,
                                                         filt_min,
                                                         filt_max,
                                                         self.num_filters,
                                                         self.ratio,
                                                         self.f.wavelen_step)
                            if filt_input is True:
                                x.append(point)
                                pts_needed -= 1
                            else:
                                opt.tell(point, 0)
                            pts_tried += 1

                        if ((pts_tried >= max_search_factor*procs) and
                           (pts_needed > 0)):
                            while pts_needed > 0:
                                best_xi = opt.Xi[np.argmin(opt.yi)]
                                next_pt = rand_state.normal(0.,
                                                            0.02*(filt_max -
                                                                  filt_min),
                                                            size=len(best_xi))
                                print(best_xi, next_pt)
                                next_pt += best_xi
                                filt_input = self.validate_filter_input(
                                    next_pt, filt_min, filt_max,
                                    self.num_filters, self.ratio,
                                    self.f.wavelen_step)
                                if filt_input is True:
                                    x.append(list(np.sort(next_pt)))
                                    random_points_used += 1
                                    pts_needed -= 1
                        print(pts_tried)
                    print("Random Points Used: %i" % random_points_used)

                y = parallel(delayed(unwrap_self_f)(arg1, val) for
                             arg1, val in zip([self]*len(x), x))

                opt.tell(x, y)

                non_zero = np.where(np.array(y) != 0)[0]
                i += len(non_zero)

                if checkpointing is True:
                    keep_rows = np.where(np.array(opt.yi) != 0)
                    np.savetxt('yi.out', np.array(opt.yi)[keep_rows])
                    np.savetxt('Xi.out', np.array(opt.Xi)[keep_rows])

                print(min(opt.yi), i)

                # Add random point information
                opt.random_pts_used = random_points_used

        if save_optimizer is not None:
            f = open(save_optimizer, 'wb')
            pickle.dump(opt, f)
            f.close()

        return opt

    def set_filters(self, filt_params):

        if self.ratio is not None:

            filt_input = []

            for i in range(self.num_filters):
                edges = np.array(filt_params[2*i:2*(i+1)])
                bottom_len = edges[1] - edges[0]
                top_len = self.ratio*bottom_len
                center = edges[0] + bottom_len/2.
                top_left = center - top_len/2.
                top_right = center + top_len/2.
                filt_input.append([edges[0], top_left, top_right, edges[1]])
        else:
            filt_input = [filt_params[4*i:4*(i+1)]
                          for i in range(self.num_filters)]

        filt_dict = self.f.trap_filters(filt_input)

        if self.frozen_filt_dict is None:
            return filt_dict
        else:
            filter_wavelengths = self.frozen_eff_lambda +\
                self.find_filt_centers(filt_input)
            filter_names_unsort = self.frozen_filt_dict.keys() +\
                filt_dict.keys()
            filter_list_unsort = self.frozen_filt_dict.values() +\
                filt_dict.values()
            if len(filter_wavelengths) != (self.num_filters +
                                           len(self.frozen_eff_lambda)):
                raise ValueError("Make sure frozen_filt_eff_wavelen is a list")
            sort_idx = np.argsort(filter_wavelengths)
            filter_names = [filter_names_unsort[idx] for idx in sort_idx]
            filter_list = [filter_list_unsort[idx] for idx in sort_idx]

            return BandpassDict(filter_list, filter_names)

    def calc_results(self, filt_params):

        filt_dict = self.set_filters(filt_params)

        if filt_dict == 0:
            return 0

        c = calcIG(filt_dict, self.shift_seds, self.z_probs,
                   sky_mag=self.sky_mag, sed_mags=self.sed_mags,
                   ref_filter=self.calib_filter)
        step_result = c.calc_IG(rand_state=np.random.RandomState(
                        np.int(np.sum(filt_params))))
        if self.verbosity >= 10:
            print(filt_params, step_result)

        return -1.*step_result



from builtins import object
import numpy

__all__ = ["PhotometricParameters"]


class DefaultPhotometricParameters(object):
    """
    This class will just contain a bunch of dict which store
    the default PhotometricParameters for LSST Bandpasses

    Users should not access this class (which is why it is
    not included in the __all__ declaration for this file).

    It is only used to initialize PhotometricParameters off of
    a bandpass name.
    """

    # Obviously, some of these parameters (effarea, gain, platescale,
    # darkcurrent, and readnoise) will not change as a function of bandpass;
    # we are just making them dicts here to be consistent with
    # everything else (and to make it possible for
    # PhotometricParameters to access them using the bandpass name
    # passed to its constructor)
    #
    # Note: all dicts contain an 'any' key which will be the default
    # value if an unknown bandpass is asked for
    #
    # 'any' values should be kept consistent with r band

    bandpassNames = ['u', 'g', 'r', 'i', 'z', 'y', 'any']

    def makeDict(value,
                 bandpassNames = ('u', 'g', 'r', 'i', 'z', 'y', 'any')):
        newdict = {}
        for f in bandpassNames:
            newdict[f] = value
        return newdict

    # exposure time in seconds
    exptimeSec = 15.0
    exptime = makeDict(exptimeSec)

    # number of exposures
    nexpN = 2
    nexp = makeDict(nexpN)

    # effective area in cm^2
    effareaCm2 = numpy.pi * (6.423/2.*100)**2
    effarea = makeDict(effareaCm2)

    # electrons per ADU
    gainADU = 2.3
    gain = makeDict(gainADU)

    # electrons per pixel per exposure
    readnoiseE = 8.8
    readnoise = makeDict(readnoiseE)

    # electrons per pixel per second
    darkcurrentE = 0.2
    darkcurrent = makeDict(darkcurrentE)

    # electrons per pixel per exposure
    othernoiseE = 0.0
    othernoise = makeDict(othernoiseE)

    # arcseconds per pixel
    platescaleAS = 0.2
    platescale = makeDict(platescaleAS)

    # systematic squared error in magnitudes
    # see Table 14 of the SRD document
    # https://docushare.lsstcorp.org/docushare/dsweb/Get/LPM-17
    sigmaSys = {'u':0.0075, 'g':0.005, 'r':0.005, 'i':0.005, 'z':0.0075, 'y':0.0075,
                'any':0.005}


class PhotometricParameters(object):

    def __init__(self, exptime=None,
                 nexp=None,
                 effarea=None,
                 gain=None,
                 readnoise=None,
                 darkcurrent=None,
                 othernoise=None,
                 platescale=None,
                 sigmaSys=None,
                 bandpass=None):

        """
        @param [in] exptime exposure time in seconds (defaults to LSST value)

        @param [in] nexp number of exposures (defaults to LSST value)

        @param [in] effarea effective area in cm^2 (defaults to LSST value)

        @param [in] gain electrons per ADU (defaults to LSST value)

        @param [in] readnoise electrons per pixel per exposure (defaults to LSST value)

        @param [in] darkcurrent electons per pixel per second (defaults to LSST value)

        @param [in] othernoise electrons per pixel per exposure (defaults to LSST value)

        @param [in] platescale arcseconds per pixel (defaults to LSST value)

        @param [in] sigmaSys systematic error in magnitudes
        (defaults to LSST value)

        @param [in] bandpass is the name of the bandpass to which these parameters
        correspond.  If set to an LSST bandpass, the constructor will initialize
        PhotometricParameters to LSST default values for that bandpass, excepting
        any parameters that have been set by hand, i.e

        myPhotParams = PhotometricParameters(nexp=3, bandpass='u')

        will initialize a PhotometricParameters object to u bandpass defaults, except
        with 3 exposures instead of 2.

        If bandpass is left as None, other parameters will default to LSST r band
        values (except for those values set by hand).  The bandpass member variable
        of PhotometricParameters will, however, remain None.
        """

        # readnoise, darkcurrent and othernoise are measured in electrons.
        # This is taken from the specifications document LSE-30 on Docushare
        # Section 3.4.2.3 states that the total noise per pixel shall be 12.7 electrons per visit
        # which the defaults sum to (remember to multply darkcurrent by the number
        # of seconds in an exposure=15). [9 e- per 15 second exposure]

        self._exptime = None
        self._nexp = None
        self._effarea = None
        self._gain = None
        self._platescale = None
        self._sigmaSys = None
        self._readnoise = None
        self._darkcurrent = None
        self._othernoise = None


        self._bandpass = bandpass
        defaults = DefaultPhotometricParameters()

        if bandpass is None:
            bandpassKey = 'any'
            # This is so we do not set the self._bandpass member variable
            # without the user's explicit consent, but we can still access
            # default values from the PhotometricParameterDefaults
        else:
            bandpassKey = bandpass

        if bandpassKey in defaults.bandpassNames:
            self._exptime = defaults.exptime[bandpassKey]
            self._nexp = defaults.nexp[bandpassKey]
            self._effarea = defaults.effarea[bandpassKey]
            self._gain = defaults.gain[bandpassKey]
            self._platescale = defaults.platescale[bandpassKey]
            self._sigmaSys = defaults.sigmaSys[bandpassKey]
            self._readnoise = defaults.readnoise[bandpassKey]
            self._darkcurrent = defaults.darkcurrent[bandpassKey]
            self._othernoise = defaults.othernoise[bandpassKey]

        if exptime is not None:
            self._exptime = exptime

        if nexp is not None:
            self._nexp = nexp

        if effarea is not None:
            self._effarea = effarea

        if gain is not None:
            self._gain = gain

        if platescale is not None:
            self._platescale = platescale

        if sigmaSys is not None:
            self._sigmaSys = sigmaSys

        if readnoise is not None:
            self._readnoise = readnoise

        if darkcurrent is not None:
            self._darkcurrent = darkcurrent

        if othernoise is not None:
            self._othernoise = othernoise

        failureMessage = ''
        failureCt = 0

        if self._exptime is None:
            failureMessage += 'did not set exptime\n'
            failureCt += 1

        if self._nexp is None:
            failureMessage += 'did not set nexp\n'
            failureCt += 1

        if self._effarea is None:
            failureMessage += 'did not set effarea\n'
            failureCt += 1

        if self._gain is None:
            failureMessage += 'did not set gain\n'
            failureCt += 1

        if self._platescale is None:
            failureMessage += 'did not set platescale\n'
            failureCt +=1

        if self._sigmaSys is None:
            failureMessage += 'did not set sigmaSys\n'
            failureCt += 1

        if self._readnoise is None:
            failureMessage += 'did not set readnoise\n'
            failureCt += 1

        if self._darkcurrent is None:
            failureMessage += 'did not set darkcurrent\n'
            failureCt +=1

        if self._othernoise is None:
            failureMessage += 'did not set othernoise\n'
            failureCt += 1

        if failureCt>0:
            raise RuntimeError('In PhotometricParameters:\n%s' % failureMessage)



    @property
    def bandpass(self):
        """
        The name of the bandpass associated with these parameters (can be None)
        """
        return self._bandpass

    @bandpass.setter
    def bandpass(self, value):
        raise RuntimeError("You should not be setting bandpass on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")

    @property
    def exptime(self):
        """
        exposure time in seconds
        """
        return self._exptime

    @exptime.setter
    def exptime(self, value):
        raise RuntimeError("You should not be setting exptime on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def nexp(self):
        """
        number of exposures
        """
        return self._nexp

    @nexp.setter
    def nexp(self, value):
        raise RuntimeError("You should not be setting nexp on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def effarea(self):
        """
        effective area in cm^2
        """
        return self._effarea

    @effarea.setter
    def effarea(self, value):
        raise RuntimeError("You should not be setting effarea on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def gain(self):
        """
        electrons per ADU
        """
        return self._gain

    @gain.setter
    def gain(self, value):
        raise RuntimeError("You should not be setting gain on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def platescale(self):
        """
        arcseconds per pixel
        """
        return self._platescale

    @platescale.setter
    def platescale(self, value):
        raise RuntimeError("You should not be setting platescale on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def readnoise(self):
        """
        electrons per pixel per exposure
        """
        return self._readnoise

    @readnoise.setter
    def readnoise(self, value):
        raise RuntimeError("You should not be setting readnoise on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def darkcurrent(self):
        """
        electrons per pixel per second
        """
        return self._darkcurrent

    @darkcurrent.setter
    def darkcurrent(self, value):
        raise RuntimeError("You should not be setting darkcurrent on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def othernoise(self):
        """
        electrons per pixel per exposure
        """
        return self._othernoise

    @othernoise.setter
    def othernoise(self,value):
        raise RuntimeError("You should not be setting othernoise on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")


    @property
    def sigmaSys(self):
        """
        systematic error in magnitudes
        """
        return self._sigmaSys


    @sigmaSys.setter
    def sigmaSys(self, value):
        raise RuntimeError("You should not be setting sigmaSys on the fly; " +
                           "Just instantiate a new case of PhotometricParameters")



# pylint: skip-file
# Slightly modified to remove lsst_utils call. Otherwise identical to LSST
# package.

from builtins import zip
from builtins import object
import copy
import numpy
import os
from collections import OrderedDict
from .Bandpass import Bandpass
from .Sed import Sed
try:
    from lsst.utils import getPackageDir
except:
    pass

__all__ = ["BandpassDict"]

class BandpassDict(object):
    """
    This class will wrap an OrderedDict of Bandpass instantiations.
    Upon instantiation, this class's constructor will resample
    the input Bandpasses to be on the same wavelength grid (defined
    by the first input Bandpass).  The constructor will then calculate
    the 2-D phiArray for quick calculation of magnitudes in all
    Bandpasses simultaneously (see the member methods magListForSed,
    magListForSedList, fluxListForSed, fluxListForSedList).
    Note: when re-sampling the wavelength grid, it is assumed that
    the first bandpass is sampled on a uniform grid (i.e. all bandpasses
    are resampled to a grid with wavlen_min, wavelen_max determined by
    the bounds of the first bandpasses grid and with wavelen_step defined
    to be the difference between the 0th and 1st element of the first
    bandpass' wavelength grid).
    The class methods loadBandpassesFromFiles and loadTotalBandpassesFromFiles
    can be used to easily read throughput files in from disk and conver them
    into BandpassDict objects.
    """

    def __init__(self, bandpassList, bandpassNameList):
        """
        @param [in] bandpassList is a list of Bandpass instantiations
        @param [in] bandpassNameList is a list of tags to be associated
        with those Bandpasses.  These will be used as keys for the BandpassDict.
        """
        self._bandpassDict = OrderedDict()
        self._wavelen_match = None
        for bandpassName, bandpass in zip(bandpassNameList, bandpassList):

            if bandpassName in self._bandpassDict:
                raise RuntimeError("The bandpass %s occurs twice in your input " % bandpassName \
                                   + "to BandpassDict")

            self._bandpassDict[bandpassName] = copy.deepcopy(bandpass)
            if self._wavelen_match is None:
                self._wavelen_match = self._bandpassDict[bandpassName].wavelen

        dummySed = Sed()
        self._phiArray, self._wavelenStep = dummySed.setupPhiArray(list(self._bandpassDict.values()))


    def __getitem__(self, bandpass):
        return self._bandpassDict[bandpass]


    def __len__(self):
        return len(self._bandpassDict)


    def __iter__(self):
        for val in self._bandpassDict:
            yield val


    def values(self):
        """
        Returns a list of the BandpassDict's values.
        """
        return list(self._bandpassDict.values())


    def keys(self):
        """
        Returns a list of the BandpassDict's keys.
        """
        return list(self._bandpassDict.keys())


    @classmethod
    def loadBandpassesFromFiles(cls,
                                bandpassNames=['u', 'g', 'r', 'i', 'z', 'y'],
                                filedir = None,
                                bandpassRoot = 'filter_',
                                componentList = ['detector.dat', 'm1.dat', 'm2.dat', 'm3.dat',
                                                 'lens1.dat', 'lens2.dat', 'lens3.dat'],
                                atmoTransmission=None):
        """
        Load bandpass information from files into BandpassDicts.
        This method will separate the bandpasses into contributions due to instrumentations
        and contributions due to the atmosphere.
        @param [in] bandpassNames is a list of strings labeling the bandpasses
        (e.g. ['u', 'g', 'r', 'i', 'z', 'y'])
        @param [in] filedir is a string indicating the name of the directory containing the
        bandpass files
        @param [in] bandpassRoot is the root of the names of the files associated with the
        bandpasses.  This method assumes that bandpasses are stored in
        filedir/bandpassRoot_bandpassNames[i].dat
        @param [in] componentList lists the files associated with bandpasses representing
        hardware components shared by all filters
        (defaults to ['detector.dat', 'm1.dat', 'm2.dat', 'm3.dat', 'lens1.dat',
                      'lense2.dat', 'lenst3.dat']
        for LSST).  These files are also expected to be stored in filedir
        @param [in] atmoTransmission is the absolute path to the file representing the
        transmissivity of the atmosphere (defaults to baseline/atmos_std.dat in the LSST
        'throughputs' package).
        @param [out] bandpassDict is a BandpassDict containing the total
        throughput (instrumentation + atmosphere)
        @param [out] hardwareBandpassDict is a BandpassDict containing
        the throughput due to instrumentation only
        """

        if filedir is None:
            try:
                fileDir = os.path.join(getPackageDir('throughputs'), 'baseline')
            except:
                pass

        if atmoTransmission is None:
            try:
                atmoTransmission = os.path.join(getPackageDir('throughputs'),
                                              'baseline','atmos_std.dat')
            except:
                pass

        commonComponents = []
        for cc in componentList:
            commonComponents.append(os.path.join(filedir,cc))

        bandpassList = []
        hardwareBandpassList = []

        for w in bandpassNames:
            components = commonComponents + [os.path.join(filedir,"%s.dat" % (bandpassRoot +w))]
            bandpassDummy = Bandpass()
            bandpassDummy.readThroughputList(components)
            hardwareBandpassList.append(bandpassDummy)

            components += [atmoTransmission]
            bandpassDummy = Bandpass()
            bandpassDummy.readThroughputList(components)
            bandpassList.append(bandpassDummy)


        bandpassDict = cls(bandpassList, bandpassNames)
        hardwareBandpassDict = cls(hardwareBandpassList, bandpassNames)

        return bandpassDict, hardwareBandpassDict

    @classmethod
    def addSystemBandpass(cls, filter_dict,
                          bandpassDir=None,
                          components=['detector.dat', 'm1.dat', 'm2.dat',
                                      'm3.dat', 'lens1.dat',
                                      'lens2.dat', 'lens3.dat'],
                          atmosFile='atmos_std.dat'):

        """
        This is very similar to loadBandpassesFromFiles above except that
        it loads a bandpassdict instead of from file.
        """

        if bandpassDir is None:
            bandpassDir = os.path.join(os.path.dirname(__file__),
                                       '..', 'data',
                                       'lsst_baseline_throughputs')

        comp_list = [os.path.join(bandpassDir, comp) for comp in components]
        comp_atmos_list = comp_list + [os.path.join(bandpassDir, atmosFile)]

        hardwareBandpass = Bandpass()
        hardwareBandpass.readThroughputList(comp_list)
        totalBandpass = Bandpass()
        totalBandpass.readThroughputList(comp_atmos_list)

        hardware_filter_list = []
        total_filter_list = []

        for key, val in zip(filter_dict.keys(), filter_dict.values()):
            trap_filt = Bandpass()
            trap_filt_a = Bandpass()
            w, sb = val.multiplyThroughputs(hardwareBandpass.wavelen,
                                            hardwareBandpass.sb)
            w_a, sb_a = val.multiplyThroughputs(totalBandpass.wavelen,
                                                totalBandpass.sb)
            trap_filt.setBandpass(w, sb)
            hardware_filter_list.append(trap_filt)
            trap_filt_a.setBandpass(w_a, sb_a)
            total_filter_list.append(trap_filt_a)

        hardware_filt_system_dict = cls(hardware_filter_list, filter_dict.keys())
        total_filt_system_dict = cls(total_filter_list, filter_dict.keys())

        return hardware_filt_system_dict, total_filt_system_dict


    @classmethod
    def loadTotalBandpassesFromFiles(cls,
                                    bandpassNames=['u', 'g', 'r', 'i', 'z', 'y'],
                                    bandpassDir = None,
                                    bandpassRoot = 'total_'):
        """
        This will take the list of band passes named by bandpassNames and load them into
        a BandpassDict
        The bandpasses loaded this way are total bandpasses: they account for instrumental
        and atmospheric transmission.
        @param [in] bandpassNames is a list of names identifying each filter.
        Defaults to ['u', 'g', 'r', 'i', 'z', 'y']
        @param [in] bandpassDir is the name of the directory where the bandpass files are stored
        @param [in] bandpassRoot contains the first part of the bandpass file name, i.e., it is assumed
        that the bandpasses are stored in files of the type
        bandpassDir/bandpassRoot_bandpassNames[i].dat
        if we want to load bandpasses for a telescope other than LSST, we would do so
        by altering bandpassDir and bandpassRoot
        @param [out] bandpassDict is a BandpassDict containing the loaded throughputs
        """
        
        if bandpassDir is None:
            try:
                bandpassDir = os.path.join(getPackageDir('throughputs'), 'baseline')
            except:
                pass

        bandpassList = []

        for w in bandpassNames:
            bandpassDummy = Bandpass()
            bandpassDummy.readThroughput(os.path.join(bandpassDir,"%s.dat" % (bandpassRoot + w)))
            bandpassList.append(bandpassDummy)

        return cls(bandpassList, bandpassNames)


    def _magListForSed(self, sedobj, indices=None):
        """
        This is a private method which will take an sedobj which has already
        been resampled to self._wavelen_match and calculate the magnitudes
        of that object in each of the bandpasses stored in this Dict.
        The results are returned as a list.
        """

        if sedobj.wavelen is None:
            return [numpy.NaN]*len(self._bandpassDict)
        else:

            #for some reason, moving this call to flambdaTofnu()
            #to a point earlier in the
            #process results in some SEDs having 'None' for fnu.
            #
            #I looked more carefully at the documentation in Sed.py
            #Any time you update flambda in any way, fnu gets set to 'None'
            #This is to prevent the two arrays from getting out synch
            #(e.g. renormalizing flambda but forgettint to renormalize fnu)
            #
            sedobj.flambdaTofnu()

            if indices is not None:
                outputList = [numpy.NaN] * len(self._bandpassDict)
                magList = sedobj.manyMagCalc(self._phiArray, self._wavelenStep, observedBandpassInd=indices)
                for i, ix in enumerate(indices):
                    outputList[ix] = magList[i]
            else:
                outputList = sedobj.manyMagCalc(self._phiArray, self._wavelenStep)

            return outputList


    def magListForSed(self, sedobj, indices=None):
        """
        Return a list of magnitudes for a single Sed object.
        @param [in] sedobj is an Sed object.  Its wavelength grid can be arbitrary.  If necessary,
        a copy will be created and resampled onto the wavelength grid of the Bandpasses before
        magnitudes are calculated.  The original Sed will be unchanged.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate magnitudes for.  Other magnitudes will be listed as numpy.NaN (i.e. this method will
        return as many magnitudes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for magnitudes you did not actually ask for)
        @param [out] magList is a list of magnitudes in the bandpasses stored in this BandpassDict
        """

        if sedobj.wavelen is not None:

            # If the Sed's wavelength grid agrees with self._wavelen_match to one part in
            # 10^6, just use the Sed as-is.  Otherwise, copy it and resample it onto
            # self._wavelen_match
            if sedobj._needResample(wavelen_match=self._wavelen_match):
                dummySed = Sed(wavelen=sedobj.wavelen, flambda=sedobj.flambda)
                dummySed.resampleSED(force=True, wavelen_match=self._wavelen_match)
            else:
                dummySed = sedobj

            return numpy.array(self._magListForSed(dummySed, indices=indices))

        else:
            return numpy.array([numpy.NaN]*len(self._bandpassDict))


    def magDictForSed(self, sedobj, indices=None):
        """
        Return an OrderedDict of magnitudes for a single Sed object.
        The OrderedDict will be keyed off of the keys to this BandpassDict
        @param [in] sedobj is an Sed object.  Its wavelength grid can be arbitrary.  If necessary,
        a copy will be created and resampled onto the wavelength grid of the Bandpasses before
        magnitudes are calculated.  The original Sed will be unchanged.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate magnitudes for.  Other magnitudes will be listed as numpy.NaN (i.e. this method will
        return as many magnitudes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for magnitudes you did not actually ask for)
        @param [out] magDict is an OrderedDict of magnitudes in the bandpasses stored in this BandpassDict
        """

        magList = self.magListForSed(sedobj, indices=indices)

        outputDict = OrderedDict()

        for ix, bp in enumerate(self._bandpassDict.keys()):
            outputDict[bp] = magList[ix]

        return outputDict


    def magListForSedList(self, sedList, indices=None):
        """
        Return a 2-D array of magnitudes from a SedList.
        Each row will correspond to a different Sed, each column
        will correspond to a different bandpass, i.e. in the case of
        mag = myBandpassDict.magListForSedList(mySedList)
        mag[0][0] will be the magnitude of the 0th Sed in the 0th bandpass
        mag[0][1] will be the magnitude of the 0th Sed in the 1st bandpass
        mag[1][1] will be the magnitude of the 1st Sed in the 1st bandpass
        etc.
        For maximum efficiency, use the wavelenMatch keyword when loading
        SEDs into your SedList and make sure that wavelenMatch = myBandpassDict.wavelenMatch.
        That way, this method will not have to waste time resampling the Seds
        onto the wavelength grid of the BandpassDict.
        @param [in] sedList is a SedList containing the Seds
        whose magnitudes are desired.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate magnitudes for.  Other magnitudes will be listed as numpy.NaN (i.e. this method will
        return as many magnitudes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for magnitudes you did not actually ask for)
        @param [out] output_list is a 2-D numpy array containing the magnitudes
        of each Sed (the rows) in each bandpass contained in this BandpassDict
        (the columns)
        """

        one_at_a_time = False
        if sedList.wavelenMatch is None:
            one_at_a_time = True
        elif sedList[0]._needResample(wavelen_match=self._wavelen_match):
            one_at_a_time = True

        output_list = []
        if one_at_a_time:
            for sed_obj in sedList:
                sub_list = self.magListForSed(sed_obj, indices=indices)
                output_list.append(sub_list)
        else:
            # the difference between this block and the block above is that the block
            # above performs the additional check of making sure that sed_obj.wavelen
            # is equivalent to self._wavelen_match
            for sed_obj in sedList:
                sub_list = self._magListForSed(sed_obj, indices=indices)
                output_list.append(sub_list)

        return numpy.array(output_list)


    def magArrayForSedList(self, sedList, indices=None):
        """
        Return a dtyped numpy array of magnitudes from a SedList.
        The array will be keyed to the keys of this BandpassDict,
        i.e. in the case of
        mag = myBandpassDict.magArrayForSedList(mySedList)
        mag['u'][0] will be the magnitude of the 0th Sed in the 'u' bandpass
        mag['u'][1] will be the magnitude of the 1st Sed in the 'u' bandpass
        mag['z'] will be a numpy array of every Sed's magnitude in the 'z' bandpass
        etc.
        For maximum efficiency, use the wavelenMatch keyword when loading
        SEDs into your SedList and make sure that wavelenMatch = myBandpassDict.wavelenMatch.
        That way, this method will not have to waste time resampling the Seds
        onto the wavelength grid of the BandpassDict.
        @param [in] sedList is a SedList containing the Seds
        whose magnitudes are desired.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate magnitudes for.  Other magnitudes will be listed as numpy.NaN (i.e. this method will
        return as many magnitudes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for magnitudes you did not actually ask for)
        @param [out] output_array is a dtyped numpy array of magnitudes (see above).
        """

        magList = self.magListForSedList(sedList, indices=indices)

        dtype = numpy.dtype([(bp, numpy.float) for bp in self._bandpassDict.keys()])

        outputArray = numpy.array([tuple(row) for row in magList], dtype=dtype)

        return outputArray


    def _fluxListForSed(self, sedobj, indices=None):
        """
        This is a private method which will take an sedobj which has already
        been resampled to self._wavelen_match and calculate the fluxes
        of that object in each of the bandpasses stored in this Dict.
        The results are returned as a list.
        """

        if sedobj.wavelen is None:
            return [numpy.NaN]*len(self._bandpassDict)
        else:

            #for some reason, moving this call to flambdaTofnu()
            #to a point earlier in the
            #process results in some SEDs having 'None' for fnu.
            #
            #I looked more carefully at the documentation in Sed.py
            #Any time you update flambda in any way, fnu gets set to 'None'
            #This is to prevent the two arrays from getting out synch
            #(e.g. renormalizing flambda but forgettint to renormalize fnu)
            #
            sedobj.flambdaTofnu()

            if indices is not None:
                outputList = [numpy.NaN] * len(self._bandpassDict)
                magList = sedobj.manyFluxCalc(self._phiArray, self._wavelenStep, observedBandpassInd=indices)
                for i, ix in enumerate(indices):
                    outputList[ix] = magList[i]
            else:
                outputList = sedobj.manyFluxCalc(self._phiArray, self._wavelenStep)

            return outputList


    def fluxListForSed(self, sedobj, indices=None):
        """
        Return a list of Fluxes for a single Sed object.
        @param [in] sedobj is an Sed object.   Its wavelength grid can be arbitrary. If necessary,
        a copy will be created and resampled onto the wavelength grid of the Bandpasses before
        fluxes are calculated.  The original Sed will be unchanged.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate fluxes for.  Other fluxes will be listed as numpy.NaN (i.e. this method will
        return as many fluxes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for fluxes you did not actually ask for)
        @param [out] fluxList is a list of fluxes in the bandpasses stored in this BandpassDict
        Note on units: Fluxes calculated this way will be the flux density integrated over the
        weighted response curve of the bandpass.  See equaiton 2.1 of the LSST Science Book
        http://www.lsst.org/scientists/scibook
        """

        if sedobj.wavelen is not None:

            # If the Sed's wavelength grid agrees with self._wavelen_match to one part in
            # 10^6, just use the Sed as-is.  Otherwise, copy it and resample it onto
            # self._wavelen_match
            if sedobj._needResample(wavelen_match=self._wavelen_match):
                dummySed = Sed(wavelen=sedobj.wavelen, flambda=sedobj.flambda)
                dummySed.resampleSED(force=True, wavelen_match=self._wavelen_match)
            else:
                dummySed = sedobj

            return numpy.array(self._fluxListForSed(dummySed, indices=indices))

        else:
            return numpy.array([numpy.NaN]*len(self._bandpassDict))


    def fluxDictForSed(self, sedobj, indices=None):
        """
        Return an OrderedDict of fluxes for a single Sed object.
        The OrderedDict will be keyed off of the keys for this BandpassDict
        @param [in] sedobj is an Sed object.   Its wavelength grid can be arbitrary. If necessary,
        a copy will be created and resampled onto the wavelength grid of the Bandpasses before
        fluxes are calculated.  The original Sed will be unchanged.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate fluxes for.  Other fluxes will be listed as numpy.NaN (i.e. this method will
        return as many fluxes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for fluxes you did not actually ask for)
        @param [out] fluxList is a list of fluxes in the bandpasses stored in this BandpassDict
        Note on units: Fluxes calculated this way will be the flux density integrated over the
        weighted response curve of the bandpass.  See equaiton 2.1 of the LSST Science Book
        http://www.lsst.org/scientists/scibook
        """
        fluxList = self.fluxListForSed(sedobj, indices=indices)

        outputDict = OrderedDict()

        for ix, bp in enumerate(self._bandpassDict.keys()):
            outputDict[bp] = fluxList[ix]

        return outputDict


    def fluxListForSedList(self, sedList, indices=None):
        """
        Return a 2-D array of fluxes from a SedList.
        Each row will correspond to a different Sed, each column
        will correspond to a different bandpass, i.e. in the case of
        flux = myBandpassDict.fluxListForSedList(mySedList)
        flux[0][0] will be the flux of the 0th Sed in the 0th bandpass
        flux[0][1] will be the flux of the 0th Sed in the 1st bandpass
        flux[1][1] will be the flux of the 1st Sed in the 1st bandpass
        etc.
        For maximum efficiency, use the wavelenMatch keyword when loading
        SEDs into your SedList and make sure that wavelenMatch = myBandpassDict.wavelenMatch.
        That way, this method will not have to waste time resampling the Seds
        onto the wavelength grid of the BandpassDict.
        @param [in] sedList is a SedList containing the Seds
        whose fluxes are desired.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate fluxes for.  Other fluxes will be listed as numpy.NaN (i.e. this method will
        return as many fluxes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for fluxes you did not actually ask for)
        @param [out] output_list is a 2-D numpy array containing the fluxes
        of each Sed (the rows) in each bandpass contained in this BandpassDict
        (the columns)
        Note on units: Fluxes calculated this way will be the flux density integrated over the
        weighted response curve of the bandpass.  See equaiton 2.1 of the LSST Science Book
        http://www.lsst.org/scientists/scibook
        """

        one_at_a_time = False
        if sedList.wavelenMatch is None:
            one_at_a_time = True
        elif sedList[0]._needResample(wavelen_match=self._wavelen_match):
            one_at_a_time = True

        output_list = []
        if one_at_a_time:
            for sed_obj in sedList:
                sub_list = self.fluxListForSed(sed_obj, indices=indices)
                output_list.append(sub_list)
        else:
            # the difference between this block and the block above is that the block
            # above performs the additional check of making sure that sed_obj.wavelen
            # is equivalent to self._wavelen_match
            for sed_obj in sedList:
                sub_list = self._fluxListForSed(sed_obj, indices=indices)
                output_list.append(sub_list)

        return numpy.array(output_list)


    def fluxArrayForSedList(self, sedList, indices=None):
        """
        Return a dtyped numpy array of fluxes from a SedList.
        The array will be keyed to the keys of this BandpassDict,
        i.e. in the case of
        flux = myBandpassDict.fluxArrayForSedList(mySedList)
        flux['u'][0] will be the flux of the 0th Sed in the 'u' bandpass
        flux['u'][1] will be the flux of the 1st Sed in the 'u' bandpass
        flux['z'] will be a numpy array of every Sed's flux in the 'z' bandpass
        etc.
        For maximum efficiency, use the wavelenMatch keyword when loading
        SEDs into your SedList and make sure that wavelenMatch = myBandpassDict.wavelenMatch.
        That way, this method will not have to waste time resampling the Seds
        onto the wavelength grid of the BandpassDict.
        @param [in] sedList is a SedList containing the Seds
        whose fluxes are desired.
        @param [in] indices is an optional list of indices indicating which bandpasses to actually
        calculate fluxes for.  Other fluxes will be listed as numpy.NaN (i.e. this method will
        return as many fluxes as were loaded with the loadBandpassesFromFiles methods; it will
        just return numpy.NaN for fluxes you did not actually ask for)
        @param [out] output_list is a 2-D numpy array containing the fluxes
        of each Sed (the rows) in each bandpass contained in this BandpassDict
        (the columns)
        Note on units: Fluxes calculated this way will be the flux density integrated over the
        weighted response curve of the bandpass.  See equaiton 2.1 of the LSST Science Book
        http://www.lsst.org/scientists/scibook
        """

        fluxList = self.fluxListForSedList(sedList, indices=indices)

        dtype = numpy.dtype([(bp, numpy.float) for bp in self._bandpassDict.keys()])

        outputArray = numpy.array([tuple(row) for row in fluxList], dtype=dtype)

        return outputArray


    @property
    def phiArray(self):
        """
        A 2-D numpy array storing the values of phi (see eqn 2.3 of the science
        book) for all of the bandpasses in this dict.
        """
        return self._phiArray


    @property
    def wavelenStep(self):
        """
        The step size of the wavelength grid for all of the bandpasses
        stored in this dict.
        """
        return self._wavelenStep


    @property
    def wavelenMatch(self):
        """
        The wavelength grid (in nm) on which all of the bandpass
        throughputs have been sampled.
        """
        return self._wavelen_match


import os
import numpy as np
from copy import deepcopy
from functools import reduce
from . import filters, spectra, calcIG
from .lsst_utils import BandpassDict

__all__ = ["_siggiBase"]


class _siggiBase(object):

    """
    Base class for siggi optimization.
    """

    def set_starting_points(self, x0, num_filters, filt_min, filt_max,
                            ratio=None, rand_state=None):

        if x0 is None:
            x0 = []
            add_pts = 7
        else:
            # Check that form is appropriate
            if ratio is not None:
                assert (np.shape(x0)[1] == 2*num_filters), \
                    str("Wrong shape of x0. " +
                        "Need 2*number of filters in each starting point.")
            else:
                assert (np.shape(x0)[1] == 4*num_filters), \
                    str("Wrong shape of x0. " +
                        "Need 4*number of filters in each starting point.")
            add_pts = 7 - len(x0)

        if rand_state is None:
            rand_state = np.random.RandomState()

        if ratio is not None:
            x0_len = 2*num_filters
            dim_list = [(filt_min, filt_max)
                        for n in range(x0_len)]
            cuts = 2
        else:
            x0_len = 4*num_filters
            dim_list = [(filt_min,
                         filt_max) for n in range(x0_len)]
            cuts = 4

        # Create multiple starting points
        x_starts = list(np.linspace(filt_min, filt_max,
                                    num_filters+1))
        x_full_space = np.empty(x0_len)
        for i in range(num_filters):
            x_full_space[cuts*i] = x_starts[i]
            x_full_space[cuts*i:cuts*(i+1)] = np.linspace(x_starts[i],
                                                          x_starts[i+1],
                                                          cuts)
        x0.append(list(x_full_space))

        space_length = filt_max - filt_min
        x_starts_l = list(np.linspace(filt_min,
                                      filt_max - space_length/2.,
                                      num_filters+1))
        x_half_space_l = np.empty(x0_len)
        for i in range(num_filters):
            x_half_space_l[cuts*i] = x_starts_l[i]
            x_half_space_l[cuts*i:cuts*(i+1)] = np.linspace(x_starts_l[i],
                                                            x_starts_l[i+1],
                                                            cuts)
        x0.append(list(x_half_space_l))

        x_starts_r = list(np.linspace(filt_min + space_length/2.,
                                      filt_max,
                                      num_filters+1))
        x_half_space_r = np.empty(x0_len)
        for i in range(num_filters):
            x_half_space_r[cuts*i] = x_starts_l[i]
            x_half_space_r[cuts*i:cuts*(i+1)] = np.linspace(x_starts_r[i],
                                                            x_starts_r[i+1],
                                                            cuts)
        x0.append(list(x_half_space_r))

        if add_pts > 0:
            for i in range(add_pts):
                x_random = rand_state.uniform(low=filt_min,
                                              high=filt_max,
                                              size=x0_len)
                x_random = list(np.sort(x_random))
                x0.append(x_random)

        return dim_list, x0

    def validate_filter_input(self, filt_edges, filt_min, filt_max,
                              num_filters, ratio=None,
                              wavelen_step=0.1):

        # Make sure input is correct shape
        if ratio is not None:
            assert (len(filt_edges) == num_filters*2)
        else:
            assert (len(filt_edges) == num_filters*4)

        if filt_edges[0] < filt_min:
            return False
        elif filt_edges[-1] > filt_max:
            return False

        if ratio is not None:

            filt_input = []

            for i in range(num_filters):
                edges = np.array(filt_edges[2*i:2*(i+1)])
                bottom_len = edges[1] - edges[0]
                top_len = ratio*bottom_len
                center = edges[0] + bottom_len/2.
                top_left = center - top_len/2.
                top_right = center + top_len/2.
                filt_input.append([edges[0], top_left, top_right, edges[1]])
        else:
            filt_input = [filt_edges[4*i:4*(i+1)]
                          for i in range(num_filters)]

        for filt_list in filt_input:
            filt_diffs = [filt_list[idx] - filt_list[idx-1]
                          for idx in range(1, len(filt_list))]
            filt_diffs = np.array(filt_diffs)
            if np.min(filt_diffs) < -1.e-6*wavelen_step:
                return False
            elif np.max(filt_diffs) <= 0:
                return False
            elif filt_list[-1] - filt_list[0] < 3.*wavelen_step:
                return False

        # At this point if there is only one filter we are good
        # Else we will continue and check that the filters centers
        # Are increasing uniformly to limit the parameter space
        if num_filters == 1:
            return True

        filt_centers = self.find_filt_centers(filt_input)
        print(filt_centers, filt_input)
        filt_diffs = [filt_centers[idx] - filt_centers[idx-1]
                      for idx in range(1, len(filt_centers))]
        filt_diffs = np.array(filt_diffs)

        if np.min(filt_diffs) < 0:
            return False

        return True

    def find_filt_centers(self, filter_details):

        """
        Take in the filter input corners and calculate the weighted center
        of the filter in wavelength space.

        We calculate the center by finding the point where half of the
        area under the transmission curve is to the left and half to
        the right of the given point.

        Input
        -----

        filter_details, list, (n_filters, 4)

            Each row should have the lower left, upper left, upper right
            and lower right corners of the filter in wavelength space.

        Returns
        -------

        filt_centers, list of floats

            The wavelength values of the calculated centers of the input
            filters.
        """

        if (len(np.shape(filter_details)) == 2 and
                np.shape(filter_details)[1] == 4):
            pass
        elif (len(np.shape(filter_details)) == 1 and
              np.shape(filter_details)[0] == 4):
            filter_details = np.reshape(filter_details, (1, 4))
        else:
            raise ValueError("Input should be (n_filters, 4) size array")

        filt_centers = []

        for filt in filter_details:
            a1 = (filt[1] - filt[0])/2.
            a2 = (filt[2] - filt[1])
            a3 = (filt[3] - filt[2])/2.
            half_area = (a1 + a2 + a3)/2.

            if a1 == half_area:
                filt_centers.append(filt[1])
            elif a1 > half_area:
                frac_a1 = half_area/a1
                length_ha = np.sqrt(frac_a1*(filt[1] - filt[0])**2.)
                filt_centers.append(filt[0] + length_ha)
            elif (a1+a2) > half_area:
                half_a2 = half_area - a1
                length_ha = (half_a2/a2)*(filt[2] - filt[1])
                filt_centers.append(filt[1] + length_ha)
            elif (a1+a2) == half_area:
                filt_centers.append(filt[2])
            else:
                half_a3 = half_area - (a1+a2)
                frac_a3 = half_a3/a3
                length_ha = np.sqrt((1-frac_a3)*(filt[3]-filt[2])**2.)
                filt_centers.append(filt[3] - length_ha)

        return filt_centers



# pylint: skip-file
# LSST Data Management System
# Copyright 2008, 2009, 2010, 2011, 2012 LSST Corporation.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#

"""
sed -
Class data:
wavelen (nm)
flambda (ergs/cm^s/s/nm)
fnu (Jansky)
zp  (basically translates to units of fnu = -8.9 (if Janskys) or 48.6 (ergs/cm^2/s/hz))
the name of the sed file
It is important to note the units are NANOMETERS, not ANGSTROMS. It is possible to rig this so you can
use angstroms instead of nm, but you should know what you're doing and understand the wavelength grid
limits applied here and in Bandpass.py.
Methods:
 Because of how these methods will be applied for catalog generation, (taking one base SED and then
  applying various dust extinctions and redshifts), many of the methods will either work on,
  and update self, OR they can be given a set of lambda/flambda arrays and then will return
  new versions of these arrays. In general, the methods will not explicitly set flambda or fnu to
  something you (the user) did not specify - so, for example, when calculating magnitudes (which depend on
  a wavelength/fnu gridded to match the given bandpass) the wavelength and fnu used are temporary copies
  and the object itself is not changed.
 In general, the philosophy of Sed.py is to not define the wavelength grid for the object until necessary
  (so, not until needed for the magnitude calculation or resampleSED is called). At that time the min/max/step
  wavelengths or the bandpass wavelengths are used to define a new wavelength grid for the sed object.
 When considering whether to use the internal wavelen/flambda (self) values, versus input values:
  For consistency, anytime self.wavelen/flambda is used, it will be updated if the values are changed
  (except in the special case of calculating magnitudes), and if self.wavelen/flambda is updated,
  self.fnu will be set to None. This is because many operations are typically chained together
  which alter flambda -- so it is more efficient to wait and recalculate fnu at the end, plus it
  avoids possible de-synchronization errors (flambda reflecting the addition of dust while fnu does
  not, for example). If arrays are passed into a method, they will not be altered and the arrays
  which are returned will be allocated new memory.
 Another general philosophy for Sed.py is use separate methods for items which only need to be generated once
  for several objects (such as the dust A_x, b_x arrays). This allows the user to optimize their code for
  faster operation, depending on what their requirements are (see example_SedBandpass_star.py and
  exampleSedBandpass_galaxy for examples).
Method include:
  setSED / setFlatSED / readSED_flambda / readSED_fnu -- to input information into Sed wavelen/flambda.
  getSED_flambda / getSED_fnu -- to return wavelen / flambda or fnu to the user.
  clearSED -- set everything to 0.
  synchronizeSED -- to calculate wavelen/flambda/fnu on the desired grid and calculate fnu.
  _checkUseSelf/needResample -- not expected to be useful to the user, rather intended for internal use.
  resampleSED -- primarily internal use, but may be useful to user. Resamples SED onto specified grid.
  flambdaTofnu / fnuToflambda -- conversion methods, does not affect wavelen gridding.
  redshiftSED -- redshifts the SED, optionally adding dimmingx
  setupCCMab / addCCMDust -- separated into two components, so that a_x/b_x can be reused between SEDS
if the wavelength range and grid is the same for each SED (calculate a_x/b_x with setupCCMab).
  multiplySED -- multiply two SEDS together.
  calcADU / calcMag / calcFlux -- with a Bandpass, calculate the ADU/magnitude/flux of a SED.
  calcFluxNorm / multiplyFluxNorm -- handle fluxnorm parameters (from UW LSST database) properly.
     These methods are intended to give a user an easy way to scale an SED to match an expected magnitude.
  renormalizeSED  -- intended for rescaling SEDS to a common flambda or fnu level.
  writeSED -- keep a file record of your SED.
  setPhiArray -- given a list of bandpasses, sets up the 2-d phiArray (for manyMagCalc) and dlambda value.
  manyMagCalc -- given 2-d phiArray and dlambda, this will return an array of magnitudes (in the same
order as the bandpasses) of this SED in each of those bandpasses.
"""

from __future__ import with_statement
from __future__ import print_function
from builtins import zip
from builtins import str
from builtins import range
from builtins import object
import warnings
import numpy
import sys
import time
import scipy.interpolate as interpolate
import gzip
import pickle
import os
from .PhysicalParameters import PhysicalParameters
import warnings
try:
    from lsst.utils import getPackageDir
except:
    pass

__all__ = ["Sed", "cache_LSST_seds", "read_close_Kurucz"]


_global_lsst_sed_cache = None

# a cache for ASCII files read-in by the user
_global_misc_sed_cache = None


class SedCacheError(Exception):
    pass


class sed_unpickler(pickle.Unpickler):

    _allowed_obj = (("numpy", "ndarray"),
                    ("numpy", "dtype"),
                    ("numpy.core.multiarray", "_reconstruct"))

    def find_class(self, module, name):
        allowed = False
        for _module, _name in self._allowed_obj:
            if module == _module and name == _name:
                allowed = True
                break

        if not allowed:
            raise RuntimeError("Cannot call find_class() on %s, %s with sed_unpickler " % (module, name)
                               + "this is for security reasons\n"
                               + "https://docs.python.org/3.1/library/pickle.html#pickle-restrict")

        if module == "numpy":
            if name == "ndarray":
                return getattr(numpy, name)
            elif name == "dtype":
                return getattr(numpy, name)
            else:
                raise RuntimeError("sed_unpickler not meant to load numpy.%s" % name)
        elif module == "numpy.core.multiarray":
            return getattr(numpy.core.multiarray, name)
        else:
            raise RuntimeError("sed_unpickler cannot handle module %s" % module)


def _validate_sed_cache():
    """
    Verifies that the pickled SED cache exists, is a dict, and contains
    an entry for every SED in starSED/ and galaxySED.  Does nothing if so,
    raises a RuntimeError if false.
    We are doing this here so that sims_sed_library does not have to depend
    on any lsst testing software (in which case, users would have to get
    a new copy of sims_sed_library every time the upstream software changed).
    We are doing this through a method (rather than giving users access to
    _global_lsst_sed_cache) so that users do not accidentally ruin
    _global_lsst_sed_cache.
    """
    global _global_lsst_sed_cache
    if _global_lsst_sed_cache is None:
        raise SedCacheError("_global_lsst_sed_cache does not exist")
    if not isinstance(_global_lsst_sed_cache, dict):
        raise SedCacheError("_global_lsst_sed_cache is a %s; not a dict"
                            % str(type(_global_lsst_sed_cache)))
    sed_dir = getPackageDir('sims_sed_library')
    sub_dir_list = ['galaxySED', 'starSED']
    file_ct = 0
    for sub_dir in sub_dir_list:
        tree = os.walk(os.path.join(sed_dir, sub_dir))
        for entry in tree:
            local_dir = entry[0]
            file_list = entry[2]
            for file_name in file_list:
                if file_name.endswith('.gz'):
                    full_name = os.path.join(sed_dir, sub_dir, local_dir, file_name)
                    if full_name not in _global_lsst_sed_cache:
                        raise SedCacheError("%s is not in _global_lsst_sed_cache"
                                            % full_name)
                    file_ct += 1
    if file_ct == 0:
        raise SedCacheError("There were not files in _global_lsst_sed_cache")

    return


def _compare_cached_versus_uncached():
    """
    Verify that loading an SED from the pickled cache give identical
    results to loading the same SED from ASCII
    """
    sed_dir = os.path.join(getPackageDir('sims_sed_library'),
                           'starSED', 'kurucz')

    dtype = numpy.dtype([('wavelen', float), ('flambda', float)])

    sed_name_list = os.listdir(sed_dir)
    msg = ('An SED loaded from the pickled cache is not '
           'identical to the same SED loaded from ASCII; '
           'it is possible that the pickled cache was incorrectly '
           'created in sims_sed_library\n\n'
           'Try removing the cache file (the name should hav been printed '
           'to stdout above) and re-running sims_photUtils.cache_LSST_seds()')
    for ix in range(5):
        full_name = os.path.join(sed_dir, sed_name_list[ix])
        from_np = numpy.genfromtxt(full_name, dtype=dtype)
        ss_cache = Sed()
        ss_cache.readSED_flambda(full_name)
        ss_uncache = Sed(wavelen=from_np['wavelen'],
                         flambda=from_np['flambda'],
                         name=full_name)

        if not ss_cache == ss_uncache:
            raise SedCacheError(msg)


def _generate_sed_cache(cache_dir, cache_name):
    """
    Read all of the SEDs from sims_sed_library into a dict.
    Pickle the dict and store it in
    sims_photUtils/cacheDir/lsst_sed_cache.p
    Parameters
    ----------
    cache_dir is the directory where the cache will be created
    cache_name is the name of the cache to be created
    Returns
    -------
    The dict of SEDs (keyed to their full file name)
    """
    sed_root = getPackageDir('sims_sed_library')
    dtype = numpy.dtype([('wavelen', float), ('flambda', float)])

    sub_dir_list = ['agnSED', 'flatSED', 'ssmSED', 'starSED', 'galaxySED']

    cache = {}

    total_files = 0
    for sub_dir in sub_dir_list:
        dir_tree = os.walk(os.path.join(sed_root, sub_dir))
        for sub_tree in dir_tree:
            total_files += len([name for name in sub_tree[2] if name.endswith('.gz')])

    t_start = time.time()
    print("This could take about 15 minutes.")
    print("Note: not all SED files are the same size. ")
    print("Do not expect the loading rate to be uniform.\n")

    for sub_dir in sub_dir_list:
        dir_tree = os.walk(os.path.join(sed_root, sub_dir))
        for sub_tree in dir_tree:
            dir_name = sub_tree[0]
            file_list = sub_tree[2]

            for file_name in file_list:
                if file_name.endswith('.gz'):
                    try:
                        full_name = os.path.join(dir_name, file_name)
                        data = numpy.genfromtxt(full_name, dtype=dtype)
                        cache[full_name] = (data['wavelen'], data['flambda'])
                        if len(cache) % (total_files/20) == 0:
                            if len(cache) > total_files/20:
                                sys.stdout.write('\r')
                            sys.stdout.write('loaded %d of %d files in about %.2f seconds'
                                             % (len(cache), total_files, time.time()-t_start))
                            sys.stdout.flush()
                    except:
                        pass

    print('\n')

    with open(os.path.join(cache_dir, cache_name), "wb") as file_handle:
        pickle.dump(cache, file_handle)

    print('LSST SED cache saved to:\n')
    print('%s' % os.path.join(cache_dir, cache_name))

    # record the specific sims_sed_library directory being cached so that
    # a new cache will be generated if sims_sed_library gets updated
    with open(os.path.join(cache_dir, "cache_version_%d.txt" % sys.version_info.major), "w") as file_handle:
        file_handle.write("%s %s" % (sed_root, cache_name))

    return cache


def cache_LSST_seds(wavelen_min=None, wavelen_max=None):
    """
    Read all of the SEDs in sims_sed_library into a dict.  Pickle the dict
    and store it in sims_photUtils/cacheDir/lsst_sed_cache.p for future use.
    After the file has initially been created, the next time you run this script,
    it will just use pickle to load the dict.
    Once the dict is loaded, Sed.readSED_flambda() will be able to read any
    LSST-shipped SED directly from memory, rather than using I/O to read it
    from an ASCII file stored on disk.
    Note: the dict of cached SEDs will take up about 5GB on disk.  Once loaded,
    the cache will take up about 1.5GB of memory.  The cache takes about 14 minutes
    to generate and about 51 seconds to load on a 2014 Mac Book Pro.
    Parameters (optional)
    ---------------------
    wavelen_min a float
    wavelen_max a float
    if either of these are not None, then every SED in the cache will be
    truncated to only include the wavelength range (in nm) between
    wavelen_min and wavelen_max
    """

    global _global_lsst_sed_cache
    try:
        sed_cache_dir = os.path.join(getPackageDir('sims_sed_library'), 'lsst_sed_cache_dir')
        sed_cache_name = os.path.join('lsst_sed_cache_%d.p' % sys.version_info.major)
        sed_dir = getPackageDir('sims_sed_library')

    except:
        print("You did not install sims_photUtils with the full LSST simulations "
              "stack. You cannot generate and load the cache of LSST SEDs")
        return

    if not os.path.exists(sed_cache_dir):
        os.mkdir(sed_cache_dir)

    must_generate = False
    if not os.path.exists(os.path.join(sed_cache_dir, sed_cache_name)):
        must_generate = True
    if not os.path.exists(os.path.join(sed_cache_dir, "cache_version_%d.txt" % sys.version_info.major)):
        must_generate = True
    else:
        with open(os.path.join(sed_cache_dir, "cache_version_%d.txt" % sys.version_info.major), "r") as input_file:
            lines = input_file.readlines()
            if len(lines) != 1:
                must_generate = True
            else:
                info = lines[0].split()
                if len(info) != 2:
                    must_generate = True
                elif info[0] != sed_dir:
                    must_generate = True
                elif info[1] != sed_cache_name:
                    must_generate = True

    if must_generate:
        print("\nCreating cache of LSST SEDs in:\n%s" % os.path.join(sed_cache_dir, sed_cache_name))
        cache = _generate_sed_cache(sed_cache_dir, sed_cache_name)
        _global_lsst_sed_cache = cache
    else:
        print("\nOpening cache of LSST SEDs in:\n%s" % os.path.join(sed_cache_dir, sed_cache_name))
        with open(os.path.join(sed_cache_dir, sed_cache_name), 'rb') as input_file:
            _global_lsst_sed_cache = sed_unpickler(input_file).load()

    # Now that we have generated/loaded the cache, we must run tests
    # to make sure that the cache is correctly constructed.  If these
    # fail, _global_lsst_sed_cache will be set to 'None' and the code will
    # continue running.
    try:
        _validate_sed_cache()
        _compare_cached_versus_uncached()
    except SedCacheError as ee:
        print(ee.message)
        print("Cannot use cache of LSST SEDs")
        _global_lsst_sed_cache = None
        pass

    if wavelen_min is not None or wavelen_max is not None:
        if wavelen_min is None:
            wavelen_min = 0.0
        if wavelen_max is None:
            wavelen_max = numpy.inf

        new_cache = {}
        list_of_sed_names = list(_global_lsst_sed_cache.keys())
        for file_name in list_of_sed_names:
            wav, fl = _global_lsst_sed_cache.pop(file_name)
            valid_dexes = numpy.where(numpy.logical_and(wav >= wavelen_min,
                                                        wav <= wavelen_max))
            new_cache[file_name] = (wav[valid_dexes], fl[valid_dexes])

        _global_lsst_sed_cache = new_cache

    return


class Sed(object):
    """Class for holding and utilizing spectral energy distributions (SEDs)"""
    def __init__(self, wavelen=None, flambda=None, fnu=None, badval=numpy.NaN, name=None):
        """
        Initialize sed object by giving filename or lambda/flambda array.
        Note that this does *not* regrid flambda and leaves fnu undefined.
        """
        self.fnu = None
        self.wavelen = None
        self.flambda = None
        # self.zp = -8.9  # default units, Jansky.
        self.zp = -2.5*numpy.log10(3631)
        self.name = name
        self.badval = badval

        self._physParams = PhysicalParameters()

        # If init was given data to initialize class, use it.
        if (wavelen is not None) and ((flambda is not None) or (fnu is not None)):
            if name is None:
                name = 'FromArray'
            self.setSED(wavelen, flambda=flambda, fnu=fnu, name=name)
        return

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.zp != other.zp:
            return False
        if not numpy.isnan(self.badval):
            if self.badval != other.badval:
                return False
        else:
            if not numpy.isnan(other.badval):
                return False
        if self.fnu is not None and other.fnu is None:
            return False
        if self.fnu is None and other.fnu is not None:
            return False
        if self.fnu is not None:
            try:
                numpy.testing.assert_array_equal(self.fnu, other.fnu)
            except:
                return False

        if self.flambda is None and other.flambda is not None:
            return False
        if other.flambda is not None and self.flambda is None:
            return False
        if self.flambda is not None:
            try:
                numpy.testing.assert_array_equal(self.flambda, other.flambda)
            except:
                return False

        if self.wavelen is None and other.wavelen is not None:
            return False
        if self.wavelen is not None and other.wavelen is None:
            return False
        if self.wavelen is not None:
            try:
                numpy.testing.assert_array_equal(self.wavelen, other.wavelen)
            except:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    # Methods for getters and setters.

    def setSED(self, wavelen, flambda=None, fnu=None, name='FromArray'):
        """
        Populate wavelen/flambda fields in sed by giving lambda/flambda or lambda/fnu array.
        If flambda present, this overrides fnu. Method sets fnu=None unless only fnu is given.
        Sets wavelen/flambda or wavelen/flambda/fnu over wavelength array given.
        """
        # Check wavelen array for type matches.
        if isinstance(wavelen, numpy.ndarray) is False:
            raise ValueError("Wavelength must be a numpy array")
        # Wavelen type ok - make new copy of data for self.
        self.wavelen = numpy.copy(wavelen)
        self.flambda = None
        self.fnu = None
        # Check if given flambda or fnu.
        if flambda is not None:
            # Check flambda data type and length.
            if (isinstance(flambda, numpy.ndarray) is False) or (len(flambda) != len(self.wavelen)):
                raise ValueError("Flambda must be a numpy array of same length as Wavelen.")
            # Flambda ok, make a new copy of data for self.
            self.flambda = numpy.copy(flambda)
        else:
            # Were passed fnu instead : check fnu data type and length.
            if fnu is None:
                raise ValueError("Both fnu and flambda are 'None', cannot set the SED.")
            elif (isinstance(fnu, numpy.ndarray) is False) or (len(fnu) != len(self.wavelen)):
                raise ValueError("(No Flambda) - Fnu must be numpy array of same length as Wavelen.")
            # Convert fnu to flambda.
            self.wavelen, self.flambda = self.fnuToflambda(wavelen, fnu)
        self.name = name
        return

    def setFlatSED(self, wavelen_min=None,
                   wavelen_max=None,
                   wavelen_step=None, name='Flat'):
        """
        Populate the wavelength/flambda/fnu fields in sed according to a flat fnu source.
        """
        if wavelen_min is None:
            wavelen_min = self._physParams.minwavelen

        if wavelen_max is None:
            wavelen_max = self._physParams.maxwavelen

        if wavelen_step is None:
            wavelen_step = self._physParams.wavelenstep

        self.wavelen = numpy.arange(wavelen_min, wavelen_max+wavelen_step, wavelen_step, dtype='float')
        self.fnu = numpy.ones(len(self.wavelen), dtype='float') * 3631  # jansky
        self.fnuToflambda()
        self.name = name
        return

    def readSED_flambda(self, filename, name=None):
        """
        Read a file containing [lambda Flambda] (lambda in nm) (Flambda erg/cm^2/s/nm).
        Does not resample wavelen/flambda onto grid; leave fnu=None.
        """
        global _global_lsst_sed_cache
        global _global_misc_sed_cache

        # Try to open data file.
        # ASSUME that if filename ends with '.gz' that the file is gzipped. Otherwise, regular file.
        if filename.endswith('.gz'):
            gzipped_filename = filename
            unzipped_filename = filename[:-3]
        else:
            gzipped_filename = filename + '.gz'
            unzipped_filename = filename

        cached_source = None
        if _global_lsst_sed_cache is not None:
            if gzipped_filename in _global_lsst_sed_cache:
                cached_source = _global_lsst_sed_cache[gzipped_filename]
            elif unzipped_filename in _global_lsst_sed_cache:
                cached_source = _global_lsst_sed_cache[unzipped_filename]

        if cached_source is None and _global_misc_sed_cache is not None:
            if gzipped_filename in _global_misc_sed_cache:
                cached_source = _global_misc_sed_cache[gzipped_filename]
            if unzipped_filename in _global_misc_sed_cache:
                cached_source = _global_misc_sed_cache[unzipped_filename]

        if cached_source is not None:
            sourcewavelen = numpy.copy(cached_source[0])
            sourceflambda = numpy.copy(cached_source[1])

        if cached_source is None:
            # Read source SED from file - lambda, flambda should be first two columns in the file.
            # lambda should be in nm and flambda should be in ergs/cm2/s/nm
            dtype = numpy.dtype([('wavelen', float), ('flambda', float)])
            try:
                data = numpy.genfromtxt(gzipped_filename, dtype=dtype)
            except IOError:
                try:
                    data = numpy.genfromtxt(unzipped_filename, dtype=dtype)
                except Exception as err:
                    # see
                    # http://stackoverflow.com/questions/
                    # 9157210/how-do-i-raise-the-same-exception-with-a-custom-message-in-python
                    new_args = [err.args[0] + \
                                "\n\nError reading sed file %s; " % filename \
                                + "it may not exist."]
                    for aa in err.args[1:]:
                        new_args.append(aa)
                    err.args = tuple(new_args)
                    raise

            sourcewavelen = data['wavelen']
            sourceflambda = data['flambda']

            if _global_misc_sed_cache is None:
                _global_misc_sed_cache = {}
            _global_misc_sed_cache[filename] = (numpy.copy(sourcewavelen),
                                                numpy.copy(sourceflambda))

        self.wavelen = sourcewavelen
        self.flambda = sourceflambda
        self.fnu = None
        if name is None:
            self.name = filename
        else:
            self.name = name
        return

    def readSED_fnu(self, filename, name=None):
        """
        Read a file containing [lambda Fnu] (lambda in nm) (Fnu in Jansky).
        Does not resample wavelen/fnu/flambda onto a grid; leaves fnu set.
        """
        # Try to open the data file.
        try:
            if filename.endswith('.gz'):
                f = gzip.open(filename, 'r')
            else:
                f = open(filename, 'r')
        # if the above fails, look for the file with and without the gz
        except IOError:
            try:
                if filename.endswith(".gz"):
                    f = open(filename[:-3], 'r')
                else:
                    f = gzip.open(filename+".gz", 'r')
            except IOError:
                raise IOError("The throughput file %s does not exist" % (filename))
        # Read source SED from file - lambda, fnu should be first two columns in the file.
        # lambda should be in nm and fnu should be in Jansky.
        sourcewavelen = []
        sourcefnu = []
        for line in f:
            if line.startswith("#"):
                continue
            values = line.split()
            sourcewavelen.append(float(values[0]))
            sourcefnu.append(float(values[1]))
        f.close()
        # Convert to numpy arrays.
        sourcewavelen = numpy.array(sourcewavelen)
        sourcefnu = numpy.array(sourcefnu)
        # Convert fnu to flambda
        self.fnuToflambda(sourcewavelen, sourcefnu)
        if name is None:
            self.name = filename
        else:
            self.name = name
        return

    def getSED_flambda(self):
        """
        Return copy of wavelen/flambda.
        """
        # Get new memory copies of the arrays.
        wavelen = numpy.copy(self.wavelen)
        flambda = numpy.copy(self.flambda)
        return wavelen, flambda

    def getSED_fnu(self):
        """
        Return copy of wavelen/fnu, without altering self.
        """
        wavelen = numpy.copy(self.wavelen)
        # Check if fnu currently set.
        if self.fnu is not None:
            # Get new memory copy of fnu.
            fnu = numpy.copy(self.fnu)
        else:
            # Fnu was not set .. grab copy fnu without changing self.
            wavelen, fnu = self.flambdaTofnu(self.wavelen, self.flambda)
            # Now wavelen/fnu (new mem) are gridded evenly, but self.wavelen/flambda/fnu remain unchanged.
        return wavelen, fnu

    # Methods that update or change self.

    def clearSED(self):
        """
        Reset all data in sed to None.
        """
        self.wavelen = None
        self.fnu = None
        self.flambda = None
        self.zp = -8.9
        self.name = None
        return

    def synchronizeSED(self, wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Set all wavelen/flambda/fnu values, potentially on min/max/step grid.
        Uses flambda to recalculate fnu. If wavelen min/max/step are given, resamples
        wavelength/flambda/fnu onto an even grid with these values.
        """
        # Grid wavelength/flambda/fnu if desired.
        if ((wavelen_min is not None) and (wavelen_max is not None) and (wavelen_step is not None)):
            self.resampleSED(wavelen_min=wavelen_min, wavelen_max=wavelen_max,
                             wavelen_step=wavelen_step)
        # Reset or set fnu.
        self.flambdaTofnu()
        return

    # Utilities common to several later methods.

    def _checkUseSelf(self, wavelen, flux):
        """
        Simple utility to check if should be using self's data or passed arrays.
        Also does data integrity check on wavelen/flux if not self.
        """
        update_self = False
        if (wavelen is None) or (flux is None):
            # Then one of the arrays was not passed - check if this is true for both arrays.
            if (wavelen is not None) or (flux is not None):
                # Then one of the arrays was passed - raise exception.
                raise ValueError("Must either pass *both* wavelen/flux pair, or use defaults.")
            update_self = True
        else:
            # Both of the arrays were passed in - check their validity.
            if (isinstance(wavelen, numpy.ndarray) is False) or (isinstance(flux, numpy.ndarray) is False):
                raise ValueError("Must pass wavelen/flux as numpy arrays.")
            if len(wavelen) != len(flux):
                raise ValueError("Must pass equal length wavelen/flux arrays.")
        return update_self

    def _needResample(self, wavelen_match=None, wavelen=None,
                      wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Check if wavelen or self.wavelen matches wavelen or wavelen_min/max/step grid.
        """
        # Check if should use self or passed wavelen.
        if wavelen is None:
            wavelen = self.wavelen
        # Check if wavelength arrays are equal, if wavelen_match passed.
        if wavelen_match is not None:
            if numpy.shape(wavelen_match) != numpy.shape(wavelen):
                need_regrid = True
            else:
                # check the elements to see if any vary
                need_regrid = numpy.any(abs(wavelen_match-wavelen) > 1e-10)
        else:
            need_regrid = True
            # Check if wavelen_min/max/step are set - if ==None, then return (no regridding).
            # It's possible (writeSED) to call this routine, even with no final grid in mind.
            if ((wavelen_min is None) and (wavelen_max is None) and (wavelen_step is None)):
                need_regrid = False
            else:
                # Okay, now look at comparison of wavelen to the grid.
                wavelen_max_in = wavelen[len(wavelen)-1]
                wavelen_min_in = wavelen[0]
                # First check match to minimum/maximum :
                if ((wavelen_min_in == wavelen_min) and (wavelen_max_in == wavelen_max)):
                    # Then check on step size in wavelength array.
                    stepsize = numpy.unique(numpy.diff(wavelen))
                    if (len(stepsize) == 1) and (stepsize[0] == wavelen_step):
                        need_regrid = False
        # At this point, need_grid=True unless it's proven to be False, so return value.
        return need_regrid

    def resampleSED(self, wavelen=None, flux=None, wavelen_match=None,
                    wavelen_min=None, wavelen_max=None, wavelen_step=None, force=False):
        """
        Resample flux onto grid defined by min/max/step OR another wavelength array.
        Give method wavelen/flux OR default to self.wavelen/self.flambda.
        Method either returns wavelen/flambda (if given those arrays) or updates wavelen/flambda in self.
         If updating self, resets fnu to None.
         Method will first check if resampling needs to be done or not, unless 'force' is True.
        """
        # Check if need resampling:
        if force or (self._needResample(wavelen_match=wavelen_match, wavelen=wavelen, wavelen_min=wavelen_min,
                                        wavelen_max=wavelen_max, wavelen_step=wavelen_step)):
            # Is method acting on self.wavelen/flambda or passed in wavelen/flux arrays?
            update_self = self._checkUseSelf(wavelen, flux)
            if update_self:
                wavelen = self.wavelen
                flux = self.flambda
                self.fnu = None
            # Now, on with the resampling.
            # Set up gridded wavelength or copy of wavelen array to match.
            if wavelen_match is None:
                if ((wavelen_min is None) and (wavelen_max is None) and (wavelen_step is None)):
                    raise ValueError('Must set either wavelen_match or wavelen_min/max/step.')
                wavelen_grid = numpy.arange(wavelen_min, wavelen_max+wavelen_step,
                                            wavelen_step, dtype='float')
            else:
                wavelen_grid = numpy.copy(wavelen_match)
            # Check if the wavelength range desired and the wavelength range of the object overlap.
            # If there is any non-overlap, raise warning.
            if (wavelen.max() < wavelen_grid.max()) or (wavelen.min() > wavelen_grid.min()):
                warnings.warn('There is an area of non-overlap between desired wavelength range '
                              + ' (%.2f to %.2f)' % (wavelen_grid.min(), wavelen_grid.max())
                              + 'and sed %s (%.2f to %.2f)' % (self.name, wavelen.min(), wavelen.max()))
            # Do the interpolation of wavelen/flux onto grid. (type/len failures will die here).
            if wavelen[0] > wavelen_grid[0] or wavelen[-1] < wavelen_grid[-1]:
                f = interpolate.interp1d(wavelen, flux, bounds_error=False, fill_value=numpy.NaN)
                flux_grid = f(wavelen_grid)
            else:
                flux_grid = numpy.interp(wavelen_grid, wavelen, flux)

            # Update self values if necessary.
            if update_self:
                self.wavelen = wavelen_grid
                self.flambda = flux_grid
                return
            return wavelen_grid, flux_grid
        else:  # wavelength grids already match.
            update_self = self._checkUseSelf(wavelen, flux)
            if update_self:
                return
            return wavelen, flux

    def flambdaTofnu(self, wavelen=None, flambda=None):
        """
        Convert flambda into fnu.
        This routine assumes that flambda is in ergs/cm^s/s/nm and produces fnu in Jansky.
        Can act on self or user can provide wavelen/flambda and get back wavelen/fnu.
        """
        # Change Flamda to Fnu by multiplying Flambda * lambda^2 = Fv
        # Fv dv = Fl dl .. Fv = Fl dl / dv = Fl dl / (dl*c/l/l) = Fl*l*l/c
        # Check - Is the method acting on self.wavelen/flambda/fnu or passed wavelen/flambda arrays?
        update_self = self._checkUseSelf(wavelen, flambda)
        if update_self:
            wavelen = self.wavelen
            flambda = self.flambda
            self.fnu = None
        # Now on with the calculation.
        # Calculate fnu.
        fnu = flambda * wavelen * wavelen * self._physParams.nm2m / self._physParams.lightspeed
        fnu = fnu * self._physParams.ergsetc2jansky
        # If are using/updating self, then *all* wavelen/flambda/fnu will be gridded.
        # This is so wavelen/fnu AND wavelen/flambda can be kept in sync.
        if update_self:
            self.wavelen = wavelen
            self.flambda = flambda
            self.fnu = fnu
            return
        # Return wavelen, fnu, unless updating self (then does not return).
        return wavelen, fnu

    def fnuToflambda(self, wavelen=None, fnu=None):
        """
        Convert fnu into flambda.
        Assumes fnu in units of Jansky and flambda in ergs/cm^s/s/nm.
        Can act on self or user can give wavelen/fnu and get wavelen/flambda returned.
        """
        # Fv dv = Fl dl .. Fv = Fl dl / dv = Fl dl / (dl*c/l/l) = Fl*l*l/c
        # Is method acting on self or passed arrays?
        update_self = self._checkUseSelf(wavelen, fnu)
        if update_self:
            wavelen = self.wavelen
            fnu = self.fnu
        # On with the calculation.
        # Calculate flambda.
        flambda = fnu / wavelen / wavelen * self._physParams.lightspeed / self._physParams.nm2m
        flambda = flambda / self._physParams.ergsetc2jansky
        # If updating self, then *all of wavelen/fnu/flambda will be updated.
        # This is so wavelen/fnu AND wavelen/flambda can be kept in sync.
        if update_self:
            self.wavelen = wavelen
            self.flambda = flambda
            self.fnu = fnu
            return
        # Return wavelen/flambda.
        return wavelen, flambda

    # methods to alter the sed

    def redshiftSED(self, redshift, dimming=False, wavelen=None, flambda=None):
        """
        Redshift an SED, optionally adding cosmological dimming.
        Pass wavelen/flambda or redshift/update self.wavelen/flambda (unsets fnu).
        """
        # Updating self or passed arrays?
        update_self = self._checkUseSelf(wavelen, flambda)
        if update_self:
            wavelen = self.wavelen
            flambda = self.flambda
            self.fnu = None
        else:
            # Make a copy of input data, because will change its values.
            wavelen = numpy.copy(wavelen)
            flambda = numpy.copy(flambda)
        # Okay, move onto redshifting the wavelen/flambda pair.
        # Or blueshift, as the case may be.
        if redshift < 0:
            wavelen = wavelen / (1.0-redshift)
        else:
            wavelen = wavelen * (1.0+redshift)
        # Flambda now just has different wavelength for each value.
        # Add cosmological dimming if required.
        if dimming:
            if redshift < 0:
                flambda = flambda * (1.0-redshift)
            else:
                flambda = flambda / (1.0+redshift)
        # Update self, if required - but just flambda (still no grid required).
        if update_self:
            self.wavelen = wavelen
            self.flambda = flambda
            return
        return wavelen, flambda

    def setupCCMab(self, wavelen=None):
        """
        Calculate a(x) and b(x) for CCM dust model. (x=1/wavelen).
        If wavelen not specified, calculates a and b on the own object's wavelength grid.
        Returns a(x) and b(x) can be common to many seds, wavelen is the same.
        """
        # This extinction law taken from Cardelli, Clayton and Mathis ApJ 1989.
        # The general form is A_l / A(V) = a(x) + b(x)/R_V  (where x=1/lambda in microns),
        # then different values for a(x) and b(x) depending on wavelength regime.
        # Also, the extinction is parametrized as R_v = A_v / E(B-V).
        # Magnitudes of extinction (A_l) translates to flux by a_l = -2.5log(f_red / f_nonred).
        if wavelen is None:
            wavelen = numpy.copy(self.wavelen)
        a_x = numpy.zeros(len(wavelen), dtype='float')
        b_x = numpy.zeros(len(wavelen), dtype='float')
        # Convert wavelength to x (in inverse microns).
        x = numpy.empty(len(wavelen), dtype=float)
        nm_to_micron = 1/1000.0
        x = 1.0 / (wavelen * nm_to_micron)
        # Dust in infrared 0.3 /mu < x < 1.1 /mu (inverse microns).
        condition = (x >= 0.3) & (x <= 1.1)
        if len(a_x[condition]) > 0:
            y = x[condition]
            a_x[condition] = 0.574 * y**1.61
            b_x[condition] = -0.527 * y**1.61
        # Dust in optical/NIR 1.1 /mu < x < 3.3 /mu region.
        condition = (x >= 1.1) & (x <= 3.3)
        if len(a_x[condition]) > 0:
            y = x[condition] - 1.82
            a_x[condition] = 1 + 0.104*y - 0.609*y**2 + 0.701*y**3 + 1.137*y**4
            a_x[condition] = a_x[condition] - 1.718*y**5 - 0.827*y**6 + 1.647*y**7 - 0.505*y**8
            b_x[condition] = 1.952*y + 2.908*y**2 - 3.989*y**3 - 7.985*y**4
            b_x[condition] = b_x[condition] + 11.102*y**5 + 5.491*y**6 - 10.805*y**7 + 3.347*y**8
        # Dust in ultraviolet and UV (if needed for high-z) 3.3 /mu< x< 8 /mu.
        condition = (x >= 3.3) & (x < 5.9)
        if len(a_x[condition]) > 0:
            y = x[condition]
            a_x[condition] = 1.752 - 0.316*y - 0.104/((y-4.67)**2 + 0.341)
            b_x[condition] = -3.090 + 1.825*y + 1.206/((y-4.62)**2 + 0.263)
        condition = (x > 5.9) & (x < 8)
        if len(a_x[condition]) > 0:
            y = x[condition]
            Fa_x = numpy.empty(len(a_x[condition]), dtype=float)
            Fb_x = numpy.empty(len(a_x[condition]), dtype=float)
            Fa_x = -0.04473*(y-5.9)**2 - 0.009779*(y-5.9)**3
            Fb_x = 0.2130*(y-5.9)**2 + 0.1207*(y-5.9)**3
            a_x[condition] = 1.752 - 0.316*y - 0.104/((y-4.67)**2 + 0.341) + Fa_x
            b_x[condition] = -3.090 + 1.825*y + 1.206/((y-4.62)**2 + 0.263) + Fb_x
        # Dust in far UV (if needed for high-z) 8 /mu < x < 10 /mu region.
        condition = (x >= 8) & (x <= 11.)
        if len(a_x[condition]) > 0:
            y = x[condition]-8.0
            a_x[condition] = -1.073 - 0.628*(y) + 0.137*(y)**2 - 0.070*(y)**3
            b_x[condition] = 13.670 + 4.257*(y) - 0.420*(y)**2 + 0.374*(y)**3
        return a_x, b_x

    def addCCMDust(self, a_x, b_x, A_v=None, ebv=None, R_v=3.1, wavelen=None, flambda=None):
        """
        Add CCM dust model extinction to the SED, modifying flambda and fnu.
        Specify any two of A_V, E(B-V) or R_V (=3.1 default).
        """
        # The extinction law taken from Cardelli, Clayton and Mathis ApJ 1989.
        # The general form is A_l / A(V) = a(x) + b(x)/R_V  (where x=1/lambda in microns).
        # Then, different values for a(x) and b(x) depending on wavelength regime.
        # Also, the extinction is parametrized as R_v = A_v / E(B-V).
        # The magnitudes of extinction (A_l) translates to flux by a_l = -2.5log(f_red / f_nonred).
        #
        # Figure out if updating self or passed arrays.
        update_self = self._checkUseSelf(wavelen, flambda)
        if update_self:
            wavelen = self.wavelen
            flambda = self.flambda
            self.fnu = None
        else:
            wavelen = numpy.copy(wavelen)
            flambda = numpy.copy(flambda)
        # Input parameters for reddening can include any of 3 parameters; only 2 are independent.
        # Figure out what parameters were given, and see if self-consistent.
        if R_v == 3.1:
            if A_v is None:
                A_v = R_v * ebv
            elif (A_v is not None) and (ebv is not None):
                # Specified A_v and ebv, so R_v should be nondefault.
                R_v = A_v / ebv
        if (R_v != 3.1):
            if (A_v is not None) and (ebv is not None):
                calcRv = A_v / ebv
                if calcRv != R_v:
                    raise ValueError("CCM parametrization expects R_v = A_v / E(B-V);",
                                     "Please check input values, because values are inconsistent.")
            elif A_v is None:
                A_v = R_v * ebv
        # R_v and A_v values are specified or calculated.
        A_lambda = numpy.empty(len(wavelen), dtype=float)
        dust = numpy.empty(len(wavelen), dtype=float)
        A_lambda = (a_x + b_x / R_v) * A_v
        # dmag_red(dust) = -2.5 log10 (f_red / f_nored) : (f_red / f_nored) = 10**-0.4*dmag_red
        dust = numpy.power(10.0, -0.4*A_lambda)
        flambda = flambda * dust
        # Update self if required.
        if update_self:
            self.flambda = flambda
            return
        return wavelen, flambda

    def multiplySED(self, other_sed, wavelen_step=None):
        """
        Multiply two SEDs together - flambda * flambda - and return a new sed object.
        Unless the two wavelength arrays are equal, returns a SED gridded with stepsize wavelen_step
        over intersecting wavelength region. Does not alter self or other_sed.
        """

        if wavelen_step is None:
            wavelen_step = self._physParams.wavelenstep

        # Check if the wavelength arrays are equal (in which case do not resample)
        if (numpy.all(self.wavelen == other_sed.wavelen)):
            flambda = self.flambda * other_sed.flambda
            new_sed = Sed(self.wavelen, flambda=flambda)
        else:
            # Find overlapping wavelength region.
            wavelen_max = min(self.wavelen.max(), other_sed.wavelen.max())
            wavelen_min = max(self.wavelen.min(), other_sed.wavelen.min())
            if wavelen_max < wavelen_min:
                raise Exception('The two SEDS do not overlap in wavelength space.')
            # Set up wavelen/flambda of first object, on grid.
            wavelen_1, flambda_1 = self.resampleSED(self.wavelen, self.flambda,
                                                    wavelen_min=wavelen_min,
                                                    wavelen_max=wavelen_max,
                                                    wavelen_step=wavelen_step)
            # Set up wavelen/flambda of second object, on grid.
            wavelen_2, flambda_2 = self.resampleSED(wavelen=other_sed.wavelen, flux=other_sed.flambda,
                                                    wavelen_min=wavelen_min, wavelen_max=wavelen_max,
                                                    wavelen_step = wavelen_step)
            # Multiply the two flambda together.
            flambda = flambda_1 * flambda_2
            # Instantiate new sed object. wavelen_1 == wavelen_2 as both are on grid.
            new_sed = Sed(wavelen_1, flambda)
        return new_sed

    # routines related to magnitudes and fluxes

    def calcADU(self, bandpass, photParams, wavelen=None, fnu=None):
        """
        Calculate the number of adu from camera, using sb and fnu.
        Given wavelen/fnu arrays or use self. Self or passed wavelen/fnu arrays will be unchanged.
        Calculating the AB mag requires the wavelen/fnu pair to be on the same grid as bandpass;
         (temporary values of these are used).
        @param [in] bandpass is an instantiation of the Bandpass class
        @param [in] photParams is an instantiation of the
        PhotometricParameters class that carries details about the
        photometric response of the telescope.
        @param [in] wavelen (optional) is the wavelength grid in nm
        @param [in] fnu (optional) is the flux in Janskys
        If wavelen and fnu are not specified, this will just use self.wavelen and
        self.fnu
        """

        use_self = self._checkUseSelf(wavelen, fnu)
        # Use self values if desired, otherwise use values passed to function.
        if use_self:
            # Calculate fnu if required.
            if self.fnu is None:
                # If fnu not present, calculate. (does not regrid).
                self.flambdaTofnu()
            wavelen = self.wavelen
            fnu = self.fnu
        # Make sure wavelen/fnu are on the same wavelength grid as bandpass.
        wavelen, fnu = self.resampleSED(wavelen, fnu, wavelen_match=bandpass.wavelen)
        # Calculate the number of photons.
        dlambda = wavelen[1] - wavelen[0]
        # Nphoton in units of 10^-23 ergs/cm^s/nm.
        nphoton = (fnu / wavelen * bandpass.sb).sum()
        adu = nphoton * (photParams.exptime * photParams.nexp * photParams.effarea/photParams.gain) * \
              (1/self._physParams.ergsetc2jansky) * \
              (1/self._physParams.planck) * dlambda
        return adu

    def fluxFromMag(self, mag):
        """
        Convert a magnitude back into a flux (implies knowledge of the zeropoint, which is
        stored in this class)
        """

        return numpy.power(10.0, -0.4*(mag + self.zp))

    def magFromFlux(self, flux):
        """
        Convert a flux into a magnitude (implies knowledge of the zeropoint, which is stored
        in this class)
        """

        return -2.5*numpy.log10(flux) - self.zp

    def calcMag(self, bandpass, wavelen=None, fnu=None):
        """
        Calculate the AB magnitude of an object, using phi the normalized system response.
        Can pass wavelen/fnu arrays or use self. Self or passed wavelen/fnu arrays will be unchanged.
        Calculating the AB mag requires the wavelen/fnu pair to be on the same grid as bandpass;
         (but only temporary values of these are used).
         """
        # Note - the behavior in this first section might be considered a little odd.
        # However, I felt calculating a magnitude should not (unexpectedly) regrid your
        # wavelen/flambda information if you were using self., as this is not obvious from the "outside".
        # To preserve 'user logic', the wavelen/flambda of self are left untouched. Unfortunately
        # this means, this method can be used inefficiently if calculating many magnitudes with
        # the same sed and same bandpass region - in that case, use self.synchronizeSED() with
        # the wavelen min/max/step set to the bandpass min/max/step first ..
        # then you can calculate multiple magnitudes much more efficiently!
        use_self = self._checkUseSelf(wavelen, fnu)
        # Use self values if desired, otherwise use values passed to function.
        if use_self:
            # Calculate fnu if required.
            if self.fnu is None:
                self.flambdaTofnu()
            wavelen = self.wavelen
            fnu = self.fnu
        # Continue with magnitude calculation.
        # Put bandpass and wavelen/fnu are on the same grid.
        wavelen, fnu = self.resampleSED(wavelen, fnu, wavelen_match=bandpass.wavelen)
        # Calculate bandpass phi value if required.
        if bandpass.phi is None:
            bandpass.sbTophi()
        # Calculate flux in bandpass and then AB magnitude.
        dlambda = wavelen[1] - wavelen[0]
        flux = (fnu*bandpass.phi).sum() * dlambda
        if flux < 1e-300:
            raise Exception("This SED has no flux within this bandpass.")
        mag = self.magFromFlux(flux)
        return mag

    def calcFlux(self, bandpass, wavelen=None, fnu=None):
        """
        Calculate the F_b (integrated flux of an object, **above the atmosphere**), using phi.
        Passed wavelen/fnu arrays will be unchanged, but if uses self will check if fnu is set.
        Calculating the AB mag requires the wavelen/fnu pair to be on the same grid as bandpass;
           (temporary values of these are used).
        Note on units: Fluxes calculated this way will be the flux density integrated over the
        weighted response curve of the bandpass.  See equaiton 2.1 of the LSST Science Book
        http://www.lsst.org/scientists/scibook
        """
        use_self = self._checkUseSelf(wavelen, fnu)
        # Use self values if desired, otherwise use values passed to function.
        if use_self:
            # Calculate fnu if required.
            if self.fnu is None:
                self.flambdaTofnu()
            wavelen = self.wavelen
            fnu = self.fnu
        # Go on with magnitude calculation.
        wavelen, fnu = self.resampleSED(wavelen, fnu, wavelen_match=bandpass.wavelen)
        # Calculate bandpass phi value if required.
        if bandpass.phi is None:
            bandpass.sbTophi()
        # Calculate flux in bandpass and return this value.
        dlambda = wavelen[1] - wavelen[0]
        flux = (fnu*bandpass.phi).sum() * dlambda
        return flux

    def calcFluxNorm(self, magmatch, bandpass, wavelen=None, fnu=None):
        """
        Calculate the fluxNorm (SED normalization value for a given mag) for a sed.
        Equivalent to adjusting a particular f_nu to Jansky's appropriate for the desired mag.
        Can pass wavelen/fnu or apply to self.
        """
        use_self = self._checkUseSelf(wavelen, fnu)
        if use_self:
            # Check possibility that fnu is not calculated yet.
            if self.fnu is None:
                self.flambdaTofnu()
            wavelen = self.wavelen
            fnu = self.fnu
        # Fluxnorm gets applied to f_nu (fluxnorm * SED(f_nu) * PHI = mag - 8.9 (AB zeropoint).
        # FluxNorm * SED => correct magnitudes for this object.
        # Calculate fluxnorm.
        curmag = self.calcMag(bandpass, wavelen, fnu)
        if curmag == self.badval:
            return self.badval
        dmag = magmatch - curmag
        fluxnorm = numpy.power(10, (-0.4*dmag))
        return fluxnorm

    def multiplyFluxNorm(self, fluxNorm, wavelen=None, fnu=None):
        """
        Multiply wavelen/fnu (or self.wavelen/fnu) by fluxnorm.
        Returns wavelen/fnu arrays (or updates self).
        Note that multiplyFluxNorm does not regrid self.wavelen/flambda/fnu at all.
        """
        # Note that fluxNorm is intended to be applied to f_nu,
        # so that fluxnorm*fnu*phi = mag (expected magnitude).
        update_self = self._checkUseSelf(wavelen, fnu)
        if update_self:
            # Make sure fnu is defined.
            if self.fnu is None:
                self.flambdaTofnu()
            wavelen = self.wavelen
            fnu = self.fnu
        else:
            # Require new copy of the data for multiply.
            wavelen = numpy.copy(wavelen)
            fnu = numpy.copy(fnu)
        # Apply fluxnorm.
        fnu = fnu * fluxNorm
        # Update self.
        if update_self:
            self.wavelen = wavelen
            self.fnu = fnu
            # Update flambda as well.
            self.fnuToflambda()
            return
        # Else return new wavelen/fnu pairs.
        return wavelen, fnu

    def renormalizeSED(self, wavelen=None, flambda=None, fnu=None,
                       lambdanorm=500, normvalue=1, gap=0, normflux='flambda',
                       wavelen_step=None):
        """
        Renormalize sed in flambda to have normflux=normvalue @ lambdanorm or averaged over gap.
        Can normalized in flambda or fnu values. wavelen_step specifies the wavelength spacing
        when using 'gap'.
        Either returns wavelen/flambda values or updates self.
        """
        # Normalizes the fnu/flambda SED at one wavelength or average value over small range (gap).
        # This is useful for generating SED catalogs, mostly, to make them match schema.
        # Do not use this for calculating specific magnitudes -- use calcfluxNorm and multiplyFluxNorm.
        # Start normalizing wavelen/flambda.

        if wavelen_step is None:
            wavelen_step = self._physParams.wavelenstep

        if normflux == 'flambda':
            update_self = self._checkUseSelf(wavelen, flambda)
            if update_self:
                wavelen = self.wavelen
                flambda = self.flambda
            else:
                # Make a copy of the input data.
                wavelen = numpy.copy(wavelen)
                # Look for either flambda or fnu in input data.
                if flambda is None:
                    if fnu is None:
                        raise Exception("If passing wavelength, must also pass fnu or flambda.")
                    # If not given flambda, must calculate from the given values of fnu.
                    wavelen, flambda = self.fnuToflambda(wavelen, fnu)
                # Make a copy of the input data.
                else:
                    flambda = numpy.copy(flambda)
            # Calculate renormalization values.
            # Check that flambda is defined at the wavelength want to use for renormalization.
            if (lambdanorm > wavelen.max()) or (lambdanorm < wavelen.min()):
                raise Exception("Desired wavelength for renormalization, %f, " % (lambdanorm)
                                + "is outside defined wavelength range.")
            # "standard" schema have flambda = 1 at 500 nm.
            if gap == 0:
                flambda_atpt = numpy.interp(lambdanorm, wavelen, flambda, left=None, right=None)
                gapval = flambda_atpt
            else:
                lambdapt = numpy.arange(lambdanorm-gap, lambdanorm+gap, wavelen_step, dtype=float)
                flambda_atpt = numpy.zeros(len(lambdapt), dtype='float')
                flambda_atpt = numpy.interp(lambdapt, wavelen, flambda, left=None, right=None)
                gapval = flambda_atpt.sum()/len(lambdapt)
            # Now renormalize fnu and flambda in the case of normalizing flambda.
            if gapval == 0:
                raise Exception("Original flambda is 0 at the desired point of normalization. "
                                "Cannot renormalize.")
            konst = normvalue/gapval
            flambda = flambda * konst
            wavelen, fnu = self.flambdaTofnu(wavelen, flambda)
        elif normflux == 'fnu':
            update_self = self._checkUseSelf(wavelen, fnu)
            if update_self:
                wavelen = self.wavelen
                if self.fnu is None:
                    self.flambdaTofnu()
                fnu = self.fnu
            else:
                # Make a copy of the input data.
                wavelen = numpy.copy(wavelen)
                # Look for either flambda or fnu in input data.
                if fnu is None:
                    if flambda is None:
                        raise Exception("If passing wavelength, must also pass fnu or flambda.")
                    wavelen, fnu = self.flambdaTofnu(wavelen, fnu)
                # Make a copy of the input data.
                else:
                    fnu = numpy.copy(fnu)
            # Calculate renormalization values.
            # Check that flambda is defined at the wavelength want to use for renormalization.
            if (lambdanorm > wavelen.max()) or (lambdanorm < wavelen.min()):
                raise Exception("Desired wavelength for renormalization, %f, " % (lambdanorm)
                                + "is outside defined wavelength range.")
            if gap == 0:
                fnu_atpt = numpy.interp(lambdanorm, wavelen, flambda, left=None, right=None)
                gapval = fnu_atpt
            else:
                lambdapt = numpy.arange(lambdanorm-gap, lambdanorm+gap, wavelen_step, dtype=float)
                fnu_atpt = numpy.zeros(len(lambdapt), dtype='float')
                fnu_atpt = numpy.interp(lambdapt, wavelen, fnu, left=None, right=None)
                gapval = fnu_atpt.sum()/len(lambdapt)
            # Now renormalize fnu and flambda in the case of normalizing fnu.
            if gapval == 0:
                raise Exception("Original fnu is 0 at the desired point of normalization. "
                                "Cannot renormalize.")
            konst = normvalue/gapval
            fnu = fnu * konst
            wavelen, flambda = self.fnutoflambda(wavelen, fnu)
        if update_self:
            self.wavelen = wavelen
            self.flambda = flambda
            self.fnu = fnu
            return
        new_sed = Sed(wavelen=wavelen, flambda=flambda)
        return new_sed

    def writeSED(self, filename, print_header=None, print_fnu=False,
                 wavelen_min=None, wavelen_max=None, wavelen_step=None):
        """
        Write SED (wavelen, flambda, optional fnu) out to file.
        Option of adding a header line (such as version info) to output file.
        Does not alter self, regardless of grid or presence/absence of fnu.
        """
        # This can be useful for debugging or recording an SED.
        f = open(filename, 'w')
        wavelen = self.wavelen
        flambda = self.flambda
        wavelen, flambda = self.resampleSED(wavelen, flambda, wavelen_min=wavelen_min,
                                            wavelen_max=wavelen_max,
                                            wavelen_step=wavelen_step)
        # Then just use this gridded wavelen/flambda to calculate fnu.
        # Print header.
        if print_header is not None:
            if not print_header.startswith('#'):
                print_header = '# ' + print_header
            f.write(print_header)
        # Print standard header info.
        if print_fnu:
            wavelen, fnu = self.flambdaTofnu(wavelen, flambda)
            print("# Wavelength(nm)  Flambda(ergs/cm^s/s/nm)   Fnu(Jansky)", file=f)
        else:
            print("# Wavelength(nm)  Flambda(ergs/cm^s/s/nm)", file=f)
        for i in range(0, len(wavelen), 1):
            if print_fnu:
                fnu = self.flambdaTofnu(wavelen=wavelen, flambda=flambda)
                print(wavelen[i], flambda[i], fnu[i], file=f)
            else:
                print("%.2f %.7g" % (wavelen[i], flambda[i]), file=f)
        # Done writing, close file.
        f.close()
        return


# Bonus, functions for many-magnitude calculation for many SEDs with a single bandpass

    def setupPhiArray(self, bandpasslist):
        """
        Sets up a 2-d numpy phi array from bandpasslist suitable for input to Sed's manyMagCalc.
        This is intended to be used once, most likely before using Sed's manyMagCalc many times on many SEDs.
        Returns 2-d phi array and the wavelen_step (dlambda) appropriate for that array.
        """
        # Calculate dlambda for phi array.
        wavelen_step = bandpasslist[0].wavelen[1] - bandpasslist[0].wavelen[0]
        wavelen_min = bandpasslist[0].wavelen[0]
        wavelen_max = bandpasslist[0].wavelen[len(bandpasslist[0].wavelen)-1]
        # Set up
        phiarray = numpy.empty((len(bandpasslist), len(bandpasslist[0].wavelen)), dtype='float')
        # Check phis calculated and on same wavelength grid.
        i = 0
        for bp in bandpasslist:
            # Be sure bandpasses on same grid and calculate phi.
            bp.resampleBandpass(wavelen_min=wavelen_min, wavelen_max=wavelen_max, wavelen_step=wavelen_step)
            bp.sbTophi()
            phiarray[i] = bp.phi
            i = i + 1
        return phiarray, wavelen_step

    def manyFluxCalc(self, phiarray, wavelen_step, observedBandpassInd=None):
        """
        Calculate fluxes of a single sed for which fnu has been evaluated in a
        set of bandpasses for which phiarray has been set up to have the same
        wavelength grid as the SED in units of ergs/cm^2/sec. It is assumed
        that `self.fnu` is set before calling this method, and that phiArray
        has the same wavelength grid as the Sed.
        Parameters
        ----------
        phiarray: `np.ndarray`, mandatory
            phiarray corresponding to the list of bandpasses in which the band
            fluxes need to be calculated, in the same wavelength grid as the SED
        wavelen_step: `float`, mandatory
            the uniform grid size of the SED
        observedBandpassInd: list of integers, optional, defaults to None
            list of indices of phiarray corresponding to observed bandpasses,
            if None, the original phiarray is returned
        Returns
        -------
        `np.ndarray` with size equal to number of bandpass filters  band flux
        values in units of ergs/cm^2/sec
        .. note: Sed.manyFluxCalc `assumes` phiArray has the same wavelenghth
        grid as the Sed and that `sed.fnu` has been calculated for the sed,
        perhaps using `sed.flambdaTofnu()`. This requires calling
        `sed.setupPhiArray()` first. These assumptions are to avoid error
        checking within this function (for speed), but could lead to errors if
        method is used incorrectly.
        Note on units: Fluxes calculated this way will be the flux density integrated over the
        weighted response curve of the bandpass.  See equaiton 2.1 of the LSST Science Book
        http://www.lsst.org/scientists/scibook
        """

        if observedBandpassInd is not None:
            phiarray = phiarray[observedBandpassInd]
        flux = numpy.empty(len(phiarray), dtype='float')
        flux = numpy.sum(phiarray*self.fnu, axis=1)*wavelen_step
        return flux

    def manyMagCalc(self, phiarray, wavelen_step, observedBandpassInd=None):
        """
        Calculate many magnitudes for many bandpasses using a single sed.
        This method assumes that there will be flux within a particular bandpass
        (could return '-Inf' for a magnitude if there is none).
        Use setupPhiArray first, and note that Sed.manyMagCalc *assumes*
        phiArray has the same wavelength grid as the Sed, and that fnu has
        already been calculated for Sed.
        These assumptions are to avoid error checking within this function (for
        speed), but could lead to errors if method is used incorrectly.
        Parameters
        ----------
        phiarray: `np.ndarray`, mandatory
            phiarray corresponding to the list of bandpasses in which the band
            fluxes need to be calculated, in the same wavelength grid as the SED
        wavelen_step: `float`, mandatory
            the uniform grid size of the SED
        observedBandpassInd: list of integers, optional, defaults to None
            list of indices of phiarray corresponding to observed bandpasses,
            if None, the original phiarray is returned
        """
        fluxes = self.manyFluxCalc(phiarray, wavelen_step, observedBandpassInd)
        mags = -2.5*numpy.log10(fluxes) - self.zp
        return mags


def read_close_Kurucz(teff, feH, logg):
    """
    Check the cached Kurucz models and load the model closest to the input stellar parameters.
    Parameters are matched in order of Teff, feH, and logg.
    Parameters
    ----------
    teff : float
        Effective temperature of the stellar template. Reasonable range is 3830-11,100 K.
    feH : float
        Metallicity [Fe/H] of stellar template. Values in range -5 to 1.
    logg : float
       Log of the surface gravity for the stellar template. Values in range 0. to 50.
    Returns
    -------
    sed : Sed Object
        The SED of the closest matching stellar template
    paramDict : dict
        Dictionary of the teff, feH, logg that were actually loaded
    """
    global _global_lsst_sed_cache

    # Load the cache if it hasn't been done
    if _global_lsst_sed_cache is None:
        cache_LSST_seds()
    # Build an array with all the files in the cache
    if not hasattr(read_close_Kurucz, 'param_combos'):
        kurucz_files = [filename for filename
                        in _global_lsst_sed_cache if ('kurucz' in filename) &
                        ('_g' in os.path.basename(filename))]
        kurucz_files = list(set(kurucz_files))
        read_close_Kurucz.param_combos = numpy.zeros(len(kurucz_files),
                                                    dtype=[('filename', ('|U200')), ('teff', float),
                                                           ('feH', float), ('logg', float)])
        for i, filename in enumerate(kurucz_files):
            read_close_Kurucz.param_combos['filename'][i] = filename
            filename = os.path.basename(filename)
            if filename[1] == 'm':
                sign = -1
            else:
                sign = 1
            logz = sign*float(filename.split('_')[0][2:])/10.
            read_close_Kurucz.param_combos['feH'][i] = logz
            logg_temp = float(filename.split('g')[1].split('_')[0])
            read_close_Kurucz.param_combos['logg'][i] = logg_temp
            teff_temp = float(filename.split('_')[-1].split('.')[0])
            read_close_Kurucz.param_combos['teff'][i] = teff_temp
        read_close_Kurucz.param_combos = numpy.sort(read_close_Kurucz.param_combos,
                                                   order=['teff', 'feH', 'logg'])

    # Lookup the closest match. Prob a faster way to do this.
    teff_diff = numpy.abs(read_close_Kurucz.param_combos['teff'] - teff)
    g1 = numpy.where(teff_diff == teff_diff.min())[0]
    feH_diff = numpy.abs(read_close_Kurucz.param_combos['feH'][g1] - feH)
    g2 = numpy.where(feH_diff == feH_diff.min())[0]
    logg_diff = numpy.abs(read_close_Kurucz.param_combos['logg'][g1][g2] - logg)
    g3 = numpy.where(logg_diff == logg_diff.min())[0]
    fileMatch = read_close_Kurucz.param_combos['filename'][g1][g2][g3]
    if numpy.size(fileMatch > 1):
        warnings.warn('Multiple close files')
        fileMatch = fileMatch[0]

    # Record what paramters were actually loaded
    teff = read_close_Kurucz.param_combos['teff'][g1][g2][g3][0]
    feH = read_close_Kurucz.param_combos['feH'][g1][g2][g3][0]
    logg = read_close_Kurucz.param_combos['logg'][g1][g2][g3][0]

    # Read in the matching file
    sed = Sed()
    sed.readSED_flambda(fileMatch)
    return sed, {'teff': teff, 'feH': feH, 'logg': logg}



__all__ = ["calcIG", "siggi", "filters", "spectra", "integrationUtils"]

from .lsst_utils import Bandpass, BandpassDict, Sed, PhysicalParameters
from .filters import filters
from .spectra import spectra
from .mathUtils import integrationUtils
from .calcIG import calcIG
from .siggiBase import _siggiBase
from .siggi import siggi
from .plotting import plotting



import os
import numpy as np
from . import Sed

__all__ = ["spectra"]


class spectra(object):

    """
    This class will include methods to get spectra for examples.
    """

    def __init__(self):

        return

    def get_red_spectrum(self):

        sed_obj = Sed()
        sed_obj.readSED_flambda(os.path.join(os.path.dirname(__file__),
                                             'data/Inst.10E10.1Z.spec.gz'))

        return sed_obj

    def get_blue_spectrum(self):

        sed_obj = Sed()
        sed_obj.readSED_flambda(os.path.join(os.path.dirname(__file__),
                                             'data/Inst.64E08.1Z.spec.gz'))

        return sed_obj

    def get_dark_sky_spectrum(self):

        sed_obj = Sed()
        sky_sed = os.path.join(os.path.dirname(__file__),
                               'data/lsst_baseline_throughputs',
                               'darksky.dat')
        sed_obj.readSED_flambda(sky_sed)

        return sed_obj

    def get_sigmoid_spectrum(self, lam_0=364.6):

        wavelen = np.arange(9., 2400.05, 0.1)
        spec = 1 / (1 + np.exp(lam_0-wavelen))
        sed_obj = Sed()
        sed_obj.setSED(wavelen=wavelen, flambda=spec)

        return sed_obj



import numpy as np
from . import Bandpass, BandpassDict

__all__ = ["filters"]


class filters(object):

    """
    This class enables easy creation of various filters that are loaded into
    a Bandpass dictionary.
    """

    def __init__(self, wavelen_min=300., wavelen_max=1200., wavelen_step=0.1):

        self.wavelen_min = wavelen_min
        self.wavelen_max = wavelen_max
        self.wavelen_step = wavelen_step

        return

    def trap_filters(self, filter_details):

        """

        Return a bandpass dictionary with trapezoid filters.

        Input
        -----

        filter_details, list, (n_filters, 4)

            Each row should have the lower left, upper left, upper right
            and lower right corners of the filter in wavelength space.

        Returns
        -------

        bandpass_dict, BandpassDict Object

            This can then be used to calculate magnitudes on spectra
        """

        if (len(np.shape(filter_details)) == 2 and
                np.shape(filter_details)[1] == 4):
            pass
        elif (len(np.shape(filter_details)) == 1 and
              np.shape(filter_details)[0] == 4):
            filter_details = np.reshape(filter_details, (1, 4))
        else:
            raise ValueError("Input should be (n_filters, 4) size array")

        bandpass_list = []

        for band in filter_details:

            offset = self.wavelen_step / 2.

            wavelen_arr = np.arange(self.wavelen_min,
                                    self.wavelen_max+offset,
                                    self.wavelen_step)

            sb = np.zeros(len(wavelen_arr))

            climb_width_l = band[1] - band[0]
            climb_width_r = band[3] - band[2]

            min_idx = np.where(wavelen_arr >=
                               band[0])[0][0]
            min_top_idx = np.where(wavelen_arr >=
                                   band[1])[0][0]

            max_idx = np.where(wavelen_arr >
                               band[3])[0][0]
            max_top_idx = np.where(wavelen_arr >=
                                   band[2])[0][0]

            climb_steps_right = max_idx - max_top_idx
            climb_steps_left = (min_top_idx+1) - min_idx

            sb[min_idx:max_idx] = 1.0

            climb_values_left = np.linspace(0., 1.0, climb_steps_left)
            sb[min_idx:min_top_idx+1] = climb_values_left

            climb_values_right = np.linspace(1.0, 0., climb_steps_right)
            sb[max_top_idx:max_idx] = climb_values_right

            bp_object = Bandpass(wavelen=wavelen_arr, sb=sb)

            bandpass_list.append(bp_object)
            name_list = ['filter_%i' % idx for idx in
                         range(len(bandpass_list))]

            bandpass_dict = BandpassDict(bandpass_list, name_list)

        return bandpass_dict
